# pylint: disable=missing-docstring
# pylint: disable=invalid-name
# pylint: disable=unused-argument
import io
import pickle
import textwrap
import time
import unittest
from typing import (
    Dict,
    Iterator,
    Mapping,
    Optional,
    Any,
    List,
)  # pylint: disable=unused-import

import icontract
import tests.error
import tests.mock


# NOTE (mristin):
# We need to introduce a global class so that we can perform tests with pickling. Pickle module does
# not support local classes.
@icontract.invariant(lambda self: self.x > 0)
class AMadeForPickling:
    def __init__(self, x: int) -> None:
        self.x = x


@icontract.invariant(lambda self: self.x > 0)
@icontract.invariant(lambda self: self.y > 0)
class AMadeForPicklingWithSetState:
    def __init__(self, x: int) -> None:
        self.x = x

        # NOTE (mristin):
        # The attribute ``y`` is computed here, and we will not pickle it intentionally.
        self._compute_internal_state()

    def _compute_internal_state(self) -> None:
        self.y = self.x + 10

    def __getstate__(self) -> Dict[str, Any]:
        state = self.__dict__.copy()

        # NOTE (mristin):
        # We intentionally do not want to pickle ``y``.
        state.pop("y", None)

        return state

    def __setstate__(self, state: Dict[str, Any]) -> None:
        # NOTE (mristin):
        # The invariants should not be checked in __setstate__ as the object
        # will not be in the correct state.

        self.__dict__.update(state)

        # NOTE (mristin):
        # We have to re-compute the internal state.
        self._compute_internal_state()


@icontract.invariant(lambda self: self.x > 0)
@icontract.invariant(lambda self: self.y > 0)
class AMadeForPicklingWithInvalidSetState:
    def __init__(self, x: int) -> None:
        self.x = x

        # NOTE (mristin):
        # The creation does not violate the invariant.
        self.y = 1000

    def __getstate__(self) -> Dict[str, Any]:
        state = self.__dict__.copy()

        # NOTE (mristin):
        # We intentionally do not want to pickle ``y``.
        state.pop("y", None)

        return state

    def __setstate__(self, state: Dict[str, Any]) -> None:
        # NOTE (mristin):
        # The invariants should not be checked before and during the __setstate__
        # as the object will not be in the correct state, but they should be
        # checked after the call.

        self.__dict__.update(state)

        # NOTE (mristin):
        # We wrongly re-compute the internal state; this violates the invariant.
        self.y = -1000

    def __repr__(self) -> str:
        return "an instance of {}".format(self.__class__.__name__)


class TestOK(unittest.TestCase):
    def test_init(self) -> None:
        @icontract.invariant(lambda self: self.x > 0)
        class SomeClass:
            def __init__(self) -> None:
                self.x = 100

        inst = SomeClass()
        self.assertEqual(100, inst.x)

    def test_instance_method(self) -> None:
        @icontract.invariant(lambda self: self.x > 0)
        class SomeClass:
            def __init__(self) -> None:
                self.x = 100

            def some_method(self) -> None:
                self.x = 1000

        inst = SomeClass()
        inst.some_method()
        self.assertEqual(1000, inst.x)

    def test_unbound_instance_method_with_self_as_kwarg(self) -> None:
        @icontract.invariant(lambda self: self.x > 0)
        class SomeClass:
            def __init__(self) -> None:
                self.x = 100

            def some_method(self) -> None:
                self.x = 1000

        inst = SomeClass()

        func = inst.some_method.__func__  # type: ignore

        func(self=inst)

    def test_magic_method(self) -> None:
        @icontract.invariant(lambda self: self.x > 0)
        class SomeClass:
            def __init__(self) -> None:
                self.x = 100

            def __call__(self) -> None:
                self.x = 1000

        inst = SomeClass()
        inst()

        self.assertEqual(1000, inst.x)

    def test_class_method(self) -> None:
        @icontract.invariant(lambda self: self.x > 0)
        class SomeClass:
            def __init__(self) -> None:
                self.x = 100

            @classmethod
            def some_class_method(cls) -> None:
                pass

        inst = SomeClass()
        self.assertEqual(100, inst.x)

    def test_static_method(self) -> None:
        # Adapted from https://github.com/Parquery/icontract/issues/186
        @icontract.invariant(lambda self: A.some_static_method(self.x))
        @icontract.invariant(lambda self: self.some_instance_method())
        class A:
            def __init__(self) -> None:
                self.x = 10

            def some_instance_method(self) -> bool:
                # We need this instance method for easier debugging.
                return self.x < 100

            @staticmethod
            def some_static_method(x: int) -> bool:
                return x > 0

        _ = A()

    def test_inherited_static_method(self) -> None:
        @icontract.invariant(lambda self: A.some_static_method(self.x))
        @icontract.invariant(lambda self: self.some_instance_method())
        class A:
            def __init__(self) -> None:
                self.x = 10

            def some_instance_method(self) -> bool:
                # We need this instance method for easier debugging.
                return self.x < 100

            @staticmethod
            def some_static_method(x: int) -> bool:
                return x > 0

        # We need to test for inheritance.
        # See https://stackoverflow.com/questions/14187973/#comment74562120_37147128
        class B(A):
            pass

        _ = B()

    def test_protected_method_may_violate_inv(self) -> None:
        @icontract.invariant(lambda self: self.x > 0)
        class SomeClass:
            def __init__(self) -> None:
                self.x = 100

            # A protected method is allowed to break the invariant.
            def _some_protected_method(self) -> None:
                self.x = -1

            def some_method(self) -> None:
                self._some_protected_method()
                self.x = 10

        inst = SomeClass()
        inst.some_method()

        self.assertEqual(10, inst.x)

    def test_inv_broken_before_protected_method(self) -> None:
        @icontract.invariant(lambda self: self.x > 0)
        class SomeClass:
            def __init__(self) -> None:
                self.x = 100

            # A protected method can not expect the invariant to hold.
            def _some_protected_method(self) -> None:
                pass

            def some_method(self) -> None:
                self.x = -1
                self._some_protected_method()
                self.x = 10

        inst = SomeClass()
        inst.some_method()
        self.assertEqual(10, inst.x)

    def test_private_method_may_violate_inv(self) -> None:
        @icontract.invariant(lambda self: self.x > 0)
        class SomeClass:
            def __init__(self) -> None:
                self.x = 100

            # A private method is allowed to break the invariant.
            def __some_private_method(self) -> None:
                self.x = -1

            def some_method(self) -> None:
                self.__some_private_method()
                self.x = 10

        inst = SomeClass()
        inst.some_method()
        self.assertEqual(10, inst.x)

    def test_inv_broken_before_private_method(self) -> None:
        @icontract.invariant(lambda self: self.x > 0)
        class SomeClass:
            def __init__(self) -> None:
                self.x = 100

            # A private method can not expect the invariant to hold.
            def __some_private_method(self) -> None:
                pass

            def some_method(self) -> None:
                self.x = -1
                self.__some_private_method()
                self.x = 10

        inst = SomeClass()
        inst.some_method()
        self.assertEqual(10, inst.x)

    def test_inv_with_empty_arguments(self) -> None:
        z = 42

        @icontract.invariant(lambda: z == 42)
        class A:
            pass

        _ = A()

    def test_no_dict_pollution(self) -> None:
        testSelf = self

        @icontract.invariant(lambda self: self.mustHold())
        class A:
            def mustHold(self) -> bool:
                testSelf.assertDictEqual({}, self.__dict__)
                return True

        _ = A()

    def test_new_exempted(self) -> None:
        # This test is related to the issue #167.
        new_call_counter = 0
        init_call_counter = 0

        @icontract.invariant(lambda self: True)
        class Foo:
            def __new__(cls, *args, **kwargs) -> "Foo":  # type: ignore
                nonlocal new_call_counter
                new_call_counter += 1
                return super(Foo, cls).__new__(cls)

            def __init__(self) -> None:
                nonlocal init_call_counter
                init_call_counter += 1

        _ = Foo()
        self.assertEqual(1, new_call_counter)
        self.assertEqual(1, init_call_counter)

    def test_subclass_of_generic_mapping(self) -> None:
        # This test is related to the issue #167.
        counter = 0

        def increase_counter(self: Any) -> bool:
            nonlocal counter
            counter += 1
            return True

        @icontract.invariant(increase_counter)
        class Foo(Mapping[str, int]):
            def __init__(self, table: Dict[str, int]) -> None:
                self._table = table

            def __getitem__(self, key: str) -> int:
                return self._table[key]

            def __iter__(self) -> Iterator[str]:
                return iter(self._table)

            def __len__(self) -> int:
                return len(self._table)

            def __str__(self) -> str:
                return "{}({})".format(self.__class__.__name__, self._table)

        f = Foo({"a": 1})  # test the constructor
        _ = f["a"]  # test __getitem__
        _ = iter(f)  # test __iter__
        _ = len(f)  # test __len__
        _ = str(f)  # test __str__

        # 1 invariant check after the constructor +
        # 4 checks before the methods +
        # 4 checks after the methods.
        self.assertEqual(9, counter)

    def test_pickle(self) -> None:
        a = AMadeForPickling(x=2)

        buffer = io.BytesIO()
        # noinspection PyTypeChecker
        pickle.dump(a, buffer)

        buffer.seek(0)
        _ = pickle.load(buffer)

        # NOTE (mristin):
        # No invariant violation expected even though __setstate__ will result in
        # a temporarily invalid object state.

    def test_pickle_with_setstate(self) -> None:
        a = AMadeForPicklingWithSetState(x=2)

        buffer = io.BytesIO()
        # noinspection PyTypeChecker
        pickle.dump(a, buffer)

        buffer.seek(0)
        _ = pickle.load(buffer)

        # NOTE (mristin):
        # No invariant violation expected even though __setstate__ will result in
        # a temporarily invalid object state.


class TestViolation(unittest.TestCase):
    def test_init(self) -> None:
        @icontract.invariant(lambda self: self.x > 0)
        class SomeClass:
            def __init__(self, x: int) -> None:
                self.x = x

            def __repr__(self) -> str:
                return "an instance of {}".format(self.__class__.__name__)

        _ = SomeClass(x=1)

        violation_error = None  # type: Optional[icontract.ViolationError]
        try:
            _ = SomeClass(x=0)
        except icontract.ViolationError as err:
            violation_error = err

        self.assertIsNotNone(violation_error)
        self.assertEqual(
            textwrap.dedent(
                """\
                self.x > 0:
                self was an instance of SomeClass
                self.x was 0"""
            ),
            tests.error.wo_mandatory_location(str(violation_error)),
        )

    def test_inv_as_precondition(self) -> None:
        @icontract.invariant(lambda self: self.x > 0)
        class SomeClass:
            def __init__(self) -> None:
                self.x = 100

            def some_method(self) -> None:
                self.x = 10

            def __repr__(self) -> str:
                return "an instance of {}".format(self.__class__.__name__)

        violation_error = None  # type: Optional[icontract.ViolationError]
        try:
            inst = SomeClass()
            inst.x = -1
            inst.some_method()
        except icontract.ViolationError as err:
            violation_error = err

        self.assertIsNotNone(violation_error)
        self.assertEqual(
            textwrap.dedent(
                """\
                self.x > 0:
                self was an instance of SomeClass
                self.x was -1"""
            ),
            tests.error.wo_mandatory_location(str(violation_error)),
        )

    def test_method(self) -> None:
        @icontract.invariant(lambda self: self.x > 0)
        class SomeClass:
            def __init__(self) -> None:
                self.x = 100

            def some_method(self) -> None:
                self.x = -1

            def __repr__(self) -> str:
                return "an instance of {}".format(self.__class__.__name__)

        violation_error = None  # type: Optional[icontract.ViolationError]
        try:
            inst = SomeClass()
            inst.some_method()
        except icontract.ViolationError as err:
            violation_error = err

        self.assertIsNotNone(violation_error)
        self.assertEqual(
            textwrap.dedent(
                """\
                self.x > 0:
                self was an instance of SomeClass
                self.x was -1"""
            ),
            tests.error.wo_mandatory_location(str(violation_error)),
        )

    def test_magic_method(self) -> None:
        @icontract.invariant(lambda self: self.x > 0)
        class SomeClass:
            def __init__(self) -> None:
                self.x = 100

            def __call__(self) -> None:
                self.x = -1

            def __repr__(self) -> str:
                return "an instance of {}".format(self.__class__.__name__)

        violation_error = None  # type: Optional[icontract.ViolationError]
        try:
            inst = SomeClass()
            inst()
        except icontract.ViolationError as err:
            violation_error = err

        self.assertIsNotNone(violation_error)
        self.assertEqual(
            textwrap.dedent(
                """\
                self.x > 0:
                self was an instance of SomeClass
                self.x was -1"""
            ),
            tests.error.wo_mandatory_location(str(violation_error)),
        )

    def test_multiple_invs_first_violated(self) -> None:
        @icontract.invariant(lambda self: self.x > 0)
        @icontract.invariant(lambda self: self.x < 10)
        class SomeClass:
            def __init__(self) -> None:
                self.x = -1

            def __repr__(self) -> str:
                return "an instance of {}".format(self.__class__.__name__)

        violation_error = None  # type: Optional[icontract.ViolationError]
        try:
            _ = SomeClass()
        except icontract.ViolationError as err:
            violation_error = err

        self.assertIsNotNone(violation_error)
        self.assertEqual(
            textwrap.dedent(
                """\
                self.x > 0:
                self was an instance of SomeClass
                self.x was -1"""
            ),
            tests.error.wo_mandatory_location(str(violation_error)),
        )

    def test_multiple_invs_last_violated(self) -> None:
        @icontract.invariant(lambda self: self.x > 0)
        @icontract.invariant(lambda self: self.x < 10)
        class SomeClass:
            def __init__(self) -> None:
                self.x = 100

            def __repr__(self) -> str:
                return "an instance of {}".format(self.__class__.__name__)

        violation_error = None  # type: Optional[icontract.ViolationError]
        try:
            _ = SomeClass()
        except icontract.ViolationError as err:
            violation_error = err

        self.assertIsNotNone(violation_error)
        self.assertEqual(
            textwrap.dedent(
                """\
                self.x < 10:
                self was an instance of SomeClass
                self.x was 100"""
            ),
            tests.error.wo_mandatory_location(str(violation_error)),
        )

    def test_inv_violated_after_pre(self) -> None:
        @icontract.invariant(lambda self: self.x > 0)
        class SomeClass:
            def __init__(self) -> None:
                self.x = 100

            @icontract.require(lambda y: y > 0)
            def some_method(self, y: int) -> None:
                self.x = -1

            def __repr__(self) -> str:
                return "an instance of {}".format(self.__class__.__name__)

        violation_error = None  # type: Optional[icontract.ViolationError]
        try:
            inst = SomeClass()
            inst.some_method(y=-1)
        except icontract.ViolationError as err:
            violation_error = err

        self.assertIsNotNone(violation_error)
        self.assertEqual(
            textwrap.dedent(
                """\
                y > 0:
                self was an instance of SomeClass
                y was -1"""
            ),
            tests.error.wo_mandatory_location(str(violation_error)),
        )

        violation_error = None
        try:
            inst = SomeClass()
            inst.some_method(y=100)
        except icontract.ViolationError as err:
            violation_error = err

        self.assertIsNotNone(violation_error)
        self.assertEqual(
            textwrap.dedent(
                """\
                self.x > 0:
                self was an instance of SomeClass
                self.x was -1"""
            ),
            tests.error.wo_mandatory_location(str(violation_error)),
        )

    def test_inv_ok_but_post_violated(self) -> None:
        @icontract.invariant(lambda self: self.x > 0)
        class SomeClass:
            def __init__(self) -> None:
                self.x = 100

            @icontract.ensure(lambda result: result > 0)
            def some_method(self) -> int:
                self.x = 10
                return -1

            def __repr__(self) -> str:
                return "an instance of {}".format(self.__class__.__name__)

        violation_error = None  # type: Optional[icontract.ViolationError]
        try:
            inst = SomeClass()
            inst.some_method()
        except icontract.ViolationError as err:
            violation_error = err

        self.assertIsNotNone(violation_error)
        self.assertEqual(
            textwrap.dedent(
                """\
                result > 0:
                result was -1
                self was an instance of SomeClass"""
            ),
            tests.error.wo_mandatory_location(str(violation_error)),
        )

    def test_inv_violated_but_post_ok(self) -> None:
        @icontract.invariant(lambda self: self.x > 0)
        class SomeClass:
            def __init__(self) -> None:
                self.x = 100

            @icontract.ensure(lambda result: result > 0)
            def some_method(self) -> int:
                self.x = -1
                return 10

            def __repr__(self) -> str:
                return "an instance of {}".format(self.__class__.__name__)

        violation_error = None  # type: Optional[icontract.ViolationError]
        try:
            inst = SomeClass()
            inst.some_method()
        except icontract.ViolationError as err:
            violation_error = err

        self.assertIsNotNone(violation_error)
        self.assertEqual(
            textwrap.dedent(
                """\
                self.x > 0:
                self was an instance of SomeClass
                self.x was -1"""
            ),
            tests.error.wo_mandatory_location(str(violation_error)),
        )

    def test_inv_with_empty_arguments(self) -> None:
        z = 42

        @icontract.invariant(lambda: z != 42)
        class A:
            def __repr__(self) -> str:
                return "an instance of {}".format(self.__class__.__name__)

        violation_error = None  # type: Optional[icontract.ViolationError]
        try:
            _ = A()
        except icontract.ViolationError as err:
            violation_error = err

        self.assertIsNotNone(violation_error)
        self.assertEqual(
            textwrap.dedent(
                """\
                z != 42:
                self was an instance of A
                z was 42"""
            ),
            tests.error.wo_mandatory_location(str(violation_error)),
        )

    def test_condition_as_function(self) -> None:
        def some_condition(self: "A") -> bool:
            return self.x > 0

        @icontract.invariant(some_condition)
        class A:
            def __init__(self) -> None:
                self.x = 100

            def some_method(self) -> None:
                self.x = -1

            def __repr__(self) -> str:
                return "A(x={})".format(self.x)

        # Valid call
        a = A()

        # Invalid call
        violation_error = None  # type: Optional[icontract.ViolationError]
        try:
            a.some_method()
        except icontract.ViolationError as err:
            violation_error = err

        self.assertIsNotNone(violation_error)
        self.assertEqual(
            "some_condition: self was A(x=-1)",
            tests.error.wo_mandatory_location(str(violation_error)),
        )

    def test_condition_as_function_with_default_argument_value(self) -> None:
        def some_condition(self: "A", y: int = 0) -> bool:
            return self.x > y

        @icontract.invariant(some_condition)
        class A:
            def __init__(self) -> None:
                self.x = 100

            def some_method(self) -> None:
                self.x = -1

            def __repr__(self) -> str:
                return "A(x={})".format(self.x)

        # Valid call
        a = A()

        # Invalid call
        violation_error = None  # type: Optional[icontract.ViolationError]
        try:
            a.some_method()
        except icontract.ViolationError as err:
            violation_error = err

        self.assertIsNotNone(violation_error)
        self.assertEqual(
            "some_condition: self was A(x=-1)",
            tests.error.wo_mandatory_location(str(violation_error)),
        )

    def test_pickle_with_invalid_set_state(self) -> None:
        a = AMadeForPicklingWithInvalidSetState(x=2)

        buffer = io.BytesIO()
        # noinspection PyTypeChecker
        pickle.dump(a, buffer)

        buffer.seek(0)

        violation_error = None  # type: Optional[icontract.ViolationError]
        try:
            # NOTE (mristin):
            # The __setstate__ is expected to violate the invariant.
            _ = pickle.load(buffer)

        except icontract.ViolationError as err:
            violation_error = err

        self.assertIsNotNone(violation_error)
        self.assertEqual(
            """\
self.y > 0:
self was an instance of AMadeForPicklingWithInvalidSetState
self.y was -1000""",
            tests.error.wo_mandatory_location(str(violation_error)),
        )


class TestProperty(unittest.TestCase):
    def test_property_getter(self) -> None:
        @icontract.invariant(lambda self: not self.toggled)
        class SomeClass:
            def __init__(self) -> None:
                self.toggled = False

            @property
            def some_prop(self) -> int:
                self.toggled = True
                return 0

            def __repr__(self) -> str:
                return "an instance of {}".format(self.__class__.__name__)

        some_inst = SomeClass()

        violation_error = None  # type: Optional[icontract.ViolationError]
        try:
            _ = some_inst.some_prop
        except icontract.ViolationError as err:
            violation_error = err

        self.assertIsNotNone(violation_error)
        self.assertEqual(
            textwrap.dedent(
                """\
                not self.toggled:
                self was an instance of SomeClass
                self.toggled was True"""
            ),
            tests.error.wo_mandatory_location(str(violation_error)),
        )

    def test_property_setter(self) -> None:
        @icontract.invariant(lambda self: not self.toggled)
        class SomeClass:
            def __init__(self) -> None:
                self.toggled = False

            @property
            def some_prop(self) -> int:
                return 0

            @some_prop.setter
            def some_prop(self, value: int) -> None:
                self.toggled = True

            def __repr__(self) -> str:
                return "an instance of {}".format(self.__class__.__name__)

        some_inst = SomeClass()

        violation_error = None  # type: Optional[icontract.ViolationError]
        try:
            some_inst.some_prop = 0
        except icontract.ViolationError as err:
            violation_error = err

        self.assertIsNotNone(violation_error)
        self.assertEqual(
            textwrap.dedent(
                """\
                not self.toggled:
                self was an instance of SomeClass
                self.toggled was True"""
            ),
            tests.error.wo_mandatory_location(str(violation_error)),
        )

    def test_property_deleter(self) -> None:
        @icontract.invariant(lambda self: not self.toggled)
        class SomeClass:
            def __init__(self) -> None:
                self.toggled = False

            @property
            def some_prop(self) -> int:
                return 0

            @some_prop.deleter
            def some_prop(self) -> None:
                self.toggled = True

            def __repr__(self) -> str:
                return "an instance of {}".format(self.__class__.__name__)

        some_inst = SomeClass()

        violation_error = None  # type: Optional[icontract.ViolationError]
        try:
            del some_inst.some_prop
        except icontract.ViolationError as err:
            violation_error = err

        self.assertIsNotNone(violation_error)
        self.assertEqual(
            textwrap.dedent(
                """\
                not self.toggled:
                self was an instance of SomeClass
                self.toggled was True"""
            ),
            tests.error.wo_mandatory_location(str(violation_error)),
        )


class TestError(unittest.TestCase):
    def test_as_type(self) -> None:
        @icontract.invariant(lambda self: self.x > 0, error=ValueError)
        class A:
            def __init__(self) -> None:
                self.x = 0

            def __repr__(self) -> str:
                return "an instance of {}".format(self.__class__.__name__)

        value_error = None  # type: Optional[ValueError]
        try:
            _ = A()
        except ValueError as err:
            value_error = err

        self.assertIsNotNone(value_error)
        self.assertIsInstance(value_error, ValueError)
        self.assertEqual(
            textwrap.dedent(
                """\
                self.x > 0:
                self was an instance of A
                self.x was 0"""
            ),
            tests.error.wo_mandatory_location(str(value_error)),
        )

    def test_as_function(self) -> None:
        @icontract.invariant(
            lambda self: self.x > 0,
            error=lambda self: ValueError(
                "x must be positive, but got: {}".format(self.x)
            ),
        )
        class A:
            def __init__(self) -> None:
                self.x = 0

            def __repr__(self) -> str:
                return "an instance of {}".format(self.__class__.__name__)

        value_error = None  # type: Optional[ValueError]
        try:
            _ = A()
        except ValueError as err:
            value_error = err

        self.assertIsNotNone(value_error)
        self.assertIsInstance(value_error, ValueError)
        self.assertEqual("x must be positive, but got: 0", str(value_error))

    def test_as_function_with_empty_args(self) -> None:
        @icontract.invariant(
            lambda self: self.x > 0, error=lambda: ValueError("x must be positive")
        )
        class A:
            def __init__(self) -> None:
                self.x = 0

            def __repr__(self) -> str:
                return "an instance of {}".format(self.__class__.__name__)

        value_error = None  # type: Optional[ValueError]
        try:
            _ = A()
        except ValueError as err:
            value_error = err

        self.assertIsNotNone(value_error)
        self.assertIsInstance(value_error, ValueError)
        self.assertEqual("x must be positive", str(value_error))


class TestToggling(unittest.TestCase):
    def test_disabled(self) -> None:
        @icontract.invariant(lambda self: self.x > 0, enabled=False)
        class SomeClass:
            def __init__(self) -> None:
                self.x = -1

        inst = SomeClass()
        self.assertEqual(-1, inst.x)


class TestBenchmark(unittest.TestCase):
    @unittest.skip(
        "Skipped the benchmark, execute manually on a prepared benchmark machine."
    )
    def test_benchmark_when_disabled(self) -> None:
        def some_long_condition() -> bool:
            time.sleep(5)
            return True

        @icontract.invariant(lambda self: some_long_condition(), enabled=False)
        class SomeClass:
            def __init__(self) -> None:
                self.x = 100

        class AnotherClass:
            def __init__(self) -> None:
                self.x = 100

        start = time.time()
        _ = SomeClass()
        duration_with_inv = time.time() - start

        start = time.time()
        _ = AnotherClass()
        duration_wo_inv = time.time() - start

        self.assertLess(duration_with_inv / duration_wo_inv, 1.2)


class TestInvalid(unittest.TestCase):
    def test_with_invalid_arguments(self) -> None:
        val_err = None  # type: Optional[ValueError]
        try:

            @icontract.invariant(lambda self, z: self.x > z)
            class _:
                def __init__(self) -> None:
                    self.x = 100

        except ValueError as err:
            val_err = err

        self.assertIsNotNone(val_err)
        self.assertEqual(
            "Expected an invariant condition with at most an argument 'self', but got: ['self', 'z']",
            str(val_err),
        )

    def test_no_boolyness(self) -> None:
        @icontract.invariant(lambda self: tests.mock.NumpyArray([True, False]))
        class A:
            def __init__(self) -> None:
                pass

        value_error = None  # type: Optional[ValueError]
        try:
            _ = A()
        except ValueError as err:
            value_error = err

        self.assertIsNotNone(value_error)
        self.assertEqual(
            "Failed to negate the evaluation of the condition.",
            tests.error.wo_mandatory_location(str(value_error)),
        )

    # noinspection PyUnusedLocal
    def test_subclasses_affect_the_base_class_if_no_DBC(self) -> None:
        # NOTE (mristin):
        # This is a regression test for:
        # https://github.com/Parquery/icontract/issues/295
        #
        # where the invariants of the derived class erroneously "slipped into"
        # the base class when DBC or DBCMeta were not set in the ``Base`` class.

        # NOTE (mristin):
        # ``Base`` class should either inherit from ``icontract.DBC`` or have
        # the metaclass ``icontract.DBCMeta``. If it does not, the invariants will be
        # copied by reference, thus allowing the ``Derived`` class to mess them up.
        @icontract.invariant(lambda: True)
        class Base:
            def __repr__(self) -> str:
                return "an instance of {}".format(self.__class__.__name__)

        # NOTE (mristin):
        # We do not want to instantiate a derived class, but this is a regression test.
        @icontract.invariant(lambda: False)
        class Derived(Base):  # pylint: disable=unused-variable
            pass

        # NOTE (mristin):
        # This causes a ``ViolationError``, as invariants are copied by reference if
        # the ``Base`` does not inherit from ``icontract.DBC`` or uses
        # ``icontract.DBCMeta`` as the metaclass.
        violation_error = None  # type: Optional[icontract.ViolationError]
        try:
            _ = Base()
        except icontract.ViolationError as err:
            violation_error = err

        self.assertIsNotNone(violation_error)
        self.assertEqual(
            textwrap.dedent("""False: self was an instance of Base"""),
            tests.error.wo_mandatory_location(str(violation_error)),
        )


class TestCheckOn(unittest.TestCase):
    def test_invariant_checked_in_init_if_no_flag_set(self) -> None:
        # The check_on default is expected to be ``CALL`` here.
        @icontract.invariant(lambda self: self.x > 0)
        class A:
            def __init__(self) -> None:
                self.x = -1

            def __repr__(self) -> str:
                return "an instance of {}".format(self.__class__.__name__)

        violation_error = None  # type: Optional[icontract.ViolationError]
        try:
            _ = A()
        except icontract.ViolationError as err:
            violation_error = err

        self.assertIsNotNone(violation_error)
        self.assertEqual(
            textwrap.dedent(
                """\
                self.x > 0:
                self was an instance of A
                self.x was -1"""
            ),
            tests.error.wo_mandatory_location(str(violation_error)),
        )

    def test_invariant_checked_in_init_if_call_flag_set(self) -> None:
        @icontract.invariant(
            lambda self: self.x > 0, check_on=icontract.InvariantCheckEvent.CALL
        )
        class A:
            def __init__(self) -> None:
                self.x = -1

            def __repr__(self) -> str:
                return "an instance of {}".format(self.__class__.__name__)

        violation_error = None  # type: Optional[icontract.ViolationError]
        try:
            _ = A()
        except icontract.ViolationError as err:
            violation_error = err

        self.assertIsNotNone(violation_error)
        self.assertEqual(
            textwrap.dedent(
                """\
                self.x > 0:
                self was an instance of A
                self.x was -1"""
            ),
            tests.error.wo_mandatory_location(str(violation_error)),
        )

    def test_invariant_checked_in_init_if_setattr_flag_set(self) -> None:
        @icontract.invariant(
            lambda self: self.x > 0, check_on=icontract.InvariantCheckEvent.SETATTR
        )
        class A:
            def __init__(self, an_x: int) -> None:
                # NOTE (mristin):
                # Setting the ``x`` will trigger a violation here.
                self.x = an_x

            def __repr__(self) -> str:
                return "an instance of {}".format(self.__class__.__name__)

        violation_error = None  # type: Optional[icontract.ViolationError]
        try:
            _ = A(-1)
        except icontract.ViolationError as err:
            violation_error = err

        self.assertIsNotNone(violation_error)
        self.assertEqual(
            textwrap.dedent(
                """\
                self.x > 0:
                self was an instance of A
                self.x was -1"""
            ),
            tests.error.wo_mandatory_location(str(violation_error)),
        )

    def test_invariant_checked_in_init_if_all_flag_set(self) -> None:
        @icontract.invariant(
            lambda self: self.x > 0, check_on=icontract.InvariantCheckEvent.ALL
        )
        class A:
            def __init__(self) -> None:
                self.x = -1

            def __repr__(self) -> str:
                return "an instance of {}".format(self.__class__.__name__)

        violation_error = None  # type: Optional[icontract.ViolationError]
        try:
            _ = A()
        except icontract.ViolationError as err:
            violation_error = err

        self.assertIsNotNone(violation_error)
        self.assertEqual(
            textwrap.dedent(
                """\
                self.x > 0:
                self was an instance of A
                self.x was -1"""
            ),
            tests.error.wo_mandatory_location(str(violation_error)),
        )

    def test_setattr_not_affected_when_flag_not_set(self) -> None:
        # The check_on default is expected to be ``CALL`` here.
        @icontract.invariant(lambda self: self.x > 0)
        class A:
            def __init__(self) -> None:
                self.x = 10

        a = A()
        a.x = -1

        # No error expected here as we rely on the default ``check_on`` to be set to ``CALL``
        # so that setattr is not affected.

    def test_call_not_affected_when_flag_not_set(self) -> None:
        # fmt: off
        @icontract.invariant(
            lambda self: self.x > 0,
            check_on=icontract.InvariantCheckEvent.SETATTR
        )
        # fmt: on
        class A:
            def __init__(self) -> None:
                self.x = 10

            def do_something_bad(self) -> None:
                self.__dict__["x"] = -1

        a = A()
        a.do_something_bad()

        # No error expected here as ``check_on`` mandates to only check setting
        # an attribute.

    def test_setattr_affected_when_flag_set(self) -> None:
        @icontract.invariant(
            lambda self: self.x > 0, check_on=icontract.InvariantCheckEvent.SETATTR
        )
        class A:
            def __init__(self) -> None:
                self.x = 10

            def __repr__(self) -> str:
                return "an instance of {}".format(self.__class__.__name__)

        a = A()

        violation_error = None  # type: Optional[icontract.ViolationError]
        try:
            a.x = -1
        except icontract.ViolationError as err:
            violation_error = err

        self.assertIsNotNone(violation_error)
        self.assertEqual(
            textwrap.dedent(
                """\
                self.x > 0:
                self was an instance of A
                self.x was -1"""
            ),
            tests.error.wo_mandatory_location(str(violation_error)),
        )

    def test_call_and_setattr_affected_when_flags_set(self) -> None:
        @icontract.invariant(
            lambda self: self.x > 0,
            check_on=(
                icontract.InvariantCheckEvent.CALL
                | icontract.InvariantCheckEvent.SETATTR
            ),
        )
        class A:
            def __init__(self) -> None:
                self.x = 10

            def do_something_bad(self) -> None:
                self.x = -1

            def __repr__(self) -> str:
                return "an instance of {}".format(self.__class__.__name__)

        that = A()

        violation_error = None  # type: Optional[icontract.ViolationError]
        try:
            that.x = -1
        except icontract.ViolationError as err:
            violation_error = err

        self.assertIsNotNone(violation_error)
        self.assertEqual(
            textwrap.dedent(
                """\
                self.x > 0:
                self was an instance of A
                self.x was -1"""
            ),
            tests.error.wo_mandatory_location(str(violation_error)),
        )

        other = A()

        another_violation_error = None  # type: Optional[icontract.ViolationError]
        try:
            other.do_something_bad()
        except icontract.ViolationError as err:
            another_violation_error = err

        self.assertIsNotNone(another_violation_error)
        self.assertEqual(
            textwrap.dedent(
                """\
                self.x > 0:
                self was an instance of A
                self.x was -1"""
            ),
            tests.error.wo_mandatory_location(str(another_violation_error)),
        )

    def test_setattr_with_property(self) -> None:
        @icontract.invariant(
            lambda self: self.x > 0, check_on=icontract.InvariantCheckEvent.SETATTR
        )
        class A:
            def __init__(self) -> None:
                self._x = 10

            def __repr__(self) -> str:
                return "an instance of {}".format(self.__class__.__name__)

            @property
            def x(self) -> int:
                return self._x

            @x.setter
            def x(self, value: int) -> None:
                self._x = value

        a = A()

        violation_error = None  # type: Optional[icontract.ViolationError]
        try:
            a.x = -1
        except icontract.ViolationError as err:
            violation_error = err

        self.assertIsNotNone(violation_error)
        self.assertEqual(
            textwrap.dedent(
                """\
                self.x > 0:
                self was an instance of A
                self.x was -1"""
            ),
            tests.error.wo_mandatory_location(str(violation_error)),
        )

    def test_that_setattr_does_not_trigger_on_a_collection(self) -> None:
        # NOTE (mristin):
        # We want to explicitly specify that operations on collection attributes
        # do not trigger the SETATTR event.

        @icontract.invariant(
            lambda self: len(self.lst) < 2,
            check_on=icontract.InvariantCheckEvent.SETATTR,
        )
        class A:
            def __init__(self) -> None:
                self.lst = [3]

        a = A()
        # NOTE (mristin):
        # There is no ``__setattr__`` call involved here, so the invariant will not be
        # checked.
        a.lst.append(10)

    def test_that_invariant_checked_on_call_even_though_violated_through_leaking(
        self,
    ) -> None:
        # NOTE (mristin):
        # Example adapted from a comment in:
        # https://github.com/Parquery/icontract/pull/292

        @icontract.invariant(
            lambda self: all(item < 0 for item in self.lst),
            check_on=icontract.InvariantCheckEvent.CALL,
        )
        class A:
            def __init__(self, lst: List[int]) -> None:
                self.lst = lst

            def do_something(self) -> None:
                pass

            def __repr__(self) -> str:
                return "an instance of {}".format(self.__class__.__name__)

        a = A(lst=[-1])

        # NOTE (mristin):
        # The invariant will be violated here through leaking, but ``a.lst`` is not
        # changed, so no SETATTR event will be triggered.
        a.lst.append(1)

        violation_error = None  # type: Optional[icontract.ViolationError]
        try:
            # NOTE (mristin):
            # The violation must occur here as we check it *before* and *after* every method.
            a.do_something()
        except icontract.ViolationError as err:
            violation_error = err

        self.assertIsNotNone(violation_error)
        self.assertEqual(
            textwrap.dedent(
                """\
                all(item < 0 for item in self.lst):
                all(item < 0 for item in self.lst) was False, e.g., with
                  item = 1
                self was an instance of A
                self.lst was [-1, 1]"""
            ),
            tests.error.wo_mandatory_location(str(violation_error)),
        )

    def test_invariant_checked_on_setattr_if_all_flag_set(self) -> None:
        @icontract.invariant(
            lambda self: self.x > 0, check_on=icontract.InvariantCheckEvent.ALL
        )
        class A:
            def __init__(self) -> None:
                self.x = 1

            def __repr__(self) -> str:
                return "an instance of {}".format(self.__class__.__name__)

            def do_something_wrong(self) -> None:
                self.x = -1

        a = A()

        violation_error = None  # type: Optional[icontract.ViolationError]
        try:
            a.do_something_wrong()
        except icontract.ViolationError as err:
            violation_error = err

        self.assertIsNotNone(violation_error)
        self.assertEqual(
            textwrap.dedent(
                """\
                self.x > 0:
                self was an instance of A
                self.x was -1"""
            ),
            tests.error.wo_mandatory_location(str(violation_error)),
        )

    def test_invariant_checked_on_setattr_if_setattr_flag_set(self) -> None:
        @icontract.invariant(
            lambda self: self.x > 0, check_on=icontract.InvariantCheckEvent.ALL
        )
        class A:
            def __init__(self) -> None:
                self.x = 1

            def __repr__(self) -> str:
                return "an instance of {}".format(self.__class__.__name__)

        a = A()

        violation_error = None  # type: Optional[icontract.ViolationError]
        try:
            a.x = -1
        except icontract.ViolationError as err:
            violation_error = err

        self.assertIsNotNone(violation_error)
        self.assertEqual(
            textwrap.dedent(
                """\
                self.x > 0:
                self was an instance of A
                self.x was -1"""
            ),
            tests.error.wo_mandatory_location(str(violation_error)),
        )


if __name__ == "__main__":
    unittest.main()



#!/usr/bin/env python3

# pylint: disable=missing-docstring
# pylint: disable=invalid-name
# pylint: disable=unused-argument

import functools
import unittest
from typing import Optional

import icontract._checkers
from icontract._globals import CallableT


def decorator_plus_1(func: CallableT) -> CallableT:
    def wrapper(*args, **kwargs):  # type: ignore
        return func(*args, **kwargs) + 1

    functools.update_wrapper(wrapper=wrapper, wrapped=func)

    return wrapper  # type: ignore


def decorator_plus_2(func: CallableT) -> CallableT:
    def wrapper(*args, **kwargs):  # type: ignore
        return func(*args, **kwargs) + 2

    functools.update_wrapper(wrapper=wrapper, wrapped=func)

    return wrapper  # type: ignore


class TestUnwindDecoratorStack(unittest.TestCase):
    def test_wo_decorators(self) -> None:
        def func() -> int:
            return 0

        self.assertListEqual(
            [0],
            [a_func() for a_func in icontract._checkers._walk_decorator_stack(func)],
        )

    def test_with_single_decorator(self) -> None:
        @decorator_plus_1
        def func() -> int:
            return 0

        self.assertListEqual(
            [1, 0],
            [a_func() for a_func in icontract._checkers._walk_decorator_stack(func)],
        )

    def test_with_double_decorator(self) -> None:
        @decorator_plus_2
        @decorator_plus_1
        def func() -> int:
            return 0

        self.assertListEqual(
            [3, 1, 0],
            [a_func() for a_func in icontract._checkers._walk_decorator_stack(func)],
        )


class TestResolveKwargs(unittest.TestCase):
    def test_that_extra_args_raise_correct_type_error(self) -> None:
        @icontract.require(lambda: True)
        def some_func(x: int, y: int) -> None:
            pass

        type_error = None  # type: Optional[TypeError]
        try:
            some_func(1, 2, 3)  # type: ignore
        except TypeError as error:
            type_error = error

        assert type_error is not None
        self.assertRegex(
            str(type_error),
            r"^([a-zA-Z_0-9<>.]+\.)?some_func\(\) takes 2 positional arguments but 3 were given$",
        )

    def test_that_result_in_kwargs_raises_an_error(self) -> None:
        @icontract.ensure(lambda result: result > 0)
        def some_func(*args, **kwargs) -> int:  # type: ignore
            return -1

        type_error = None  # type: Optional[TypeError]

        try:
            some_func(result=-1)
        except TypeError as error:
            type_error = error

        assert type_error is not None

        self.assertEqual(
            "Unexpected argument 'result' in a function decorated with postconditions.",
            str(type_error),
        )

    def test_that_OLD_in_kwargs_raises_an_error(self) -> None:
        @icontract.ensure(lambda result: result > 0)
        def some_func(*args, **kwargs) -> int:  # type: ignore
            return -1

        type_error = None  # type: Optional[TypeError]

        try:
            some_func(OLD=-1)
        except TypeError as error:
            type_error = error

        assert type_error is not None

        self.assertEqual(
            "Unexpected argument 'OLD' in a function decorated with postconditions.",
            str(type_error),
        )


if __name__ == "__main__":
    unittest.main()



# pylint: disable=missing-docstring,unnecessary-lambda
import threading
import unittest

import icontract


@icontract.require(lambda: other_func())  # type: ignore
def some_func() -> bool:
    return True


@icontract.require(lambda: some_func())
def other_func() -> bool:
    return True


class TestThreading(unittest.TestCase):
    def test_two_threads(self) -> None:
        """
        Test that icontract can run in a multi-threaded environment.

        This is a regression test. The threading.local() can be only set
        immutable values. See
        http://slinkp.com/python-thread-locals-20171201.html for more details.

        Originally, this caused the exception
        `AttributeError: '_thread._local' object has no attribute [...]`.
        """

        class Worker(threading.Thread):
            def run(self) -> None:
                some_func()

        worker, another_worker = Worker(), Worker()
        worker.start()
        another_worker.start()
        worker.join()
        another_worker.join()


if __name__ == "__main__":
    unittest.main()



# pylint: disable=missing-docstring
# pylint: disable=invalid-name

import abc
import sys
import textwrap
import unittest
from typing import Optional  # pylint: disable=unused-import

import icontract
import tests.error


class TestOK(unittest.TestCase):
    def test_ensure_then(self) -> None:
        class A(icontract.DBC):
            @icontract.ensure(lambda result: result % 2 == 0)
            def func(self) -> int:
                return 10

        class B(A):
            @icontract.ensure(lambda result: result % 3 == 0)
            def func(self) -> int:
                return 6

        b = B()
        b.func()

    def test_count_checks(self) -> None:
        class Increment:
            count = 0

            def __call__(self) -> bool:
                Increment.count += 1
                return True

        inc = Increment()

        @icontract.invariant(lambda self: inc())
        class SomeClass:
            def __init__(self) -> None:
                pass

            def some_func(self) -> None:
                return

        inst = SomeClass()
        self.assertEqual(
            1, Increment.count
        )  # Invariant needs to be checked once after the initialization.

        inst.some_func()
        self.assertEqual(
            3, Increment.count
        )  # Invariant needs to be checked before and after some_func.

    def test_count_checks_in_slot_wrappers(self) -> None:
        class Increment:
            count = 0

            def __call__(self) -> bool:
                Increment.count += 1
                return True

        inc = Increment()

        @icontract.invariant(lambda self: inc())
        class SomeClass:
            pass

        inst = SomeClass()
        self.assertEqual(
            1, Increment.count
        )  # Invariant needs to be checked once after the initialization.

        _ = str(inst)
        self.assertEqual(
            3, Increment.count
        )  # Invariant needs to be checked before and after __str__.


class TestViolation(unittest.TestCase):
    def test_inherited_without_implementation(self) -> None:
        class A(icontract.DBC):
            @icontract.ensure(lambda result: result < 100)
            def func(self) -> int:
                return 1000

        class B(A):
            def __repr__(self) -> str:
                return "an instance of {}".format(self.__class__.__name__)

        b = B()
        violation_error = None  # type: Optional[icontract.ViolationError]
        try:
            b.func()
        except icontract.ViolationError as err:
            violation_error = err

        self.assertIsNotNone(violation_error)
        self.assertEqual(
            textwrap.dedent(
                """\
                result < 100:
                result was 1000
                self was an instance of B"""
            ),
            tests.error.wo_mandatory_location(str(violation_error)),
        )

    def test_inherited_with_modified_implementation(self) -> None:
        class A(icontract.DBC):
            @icontract.ensure(lambda result: result < 100)
            def func(self) -> int:
                return 1000

        class B(A):
            def func(self) -> int:
                return 10000

            def __repr__(self) -> str:
                return "an instance of {}".format(self.__class__.__name__)

        b = B()
        violation_error = None  # type: Optional[icontract.ViolationError]
        try:
            b.func()
        except icontract.ViolationError as err:
            violation_error = err

        self.assertIsNotNone(violation_error)
        self.assertEqual(
            textwrap.dedent(
                """\
                result < 100:
                result was 10000
                self was an instance of B"""
            ),
            tests.error.wo_mandatory_location(str(violation_error)),
        )

    def test_ensure_then_violated_in_base(self) -> None:
        class A(icontract.DBC):
            @icontract.ensure(lambda result: result % 2 == 0)
            def func(self) -> int:
                return 10

        class B(A):
            @icontract.ensure(lambda result: result % 3 == 0)
            def func(self) -> int:
                return 3

            def __repr__(self) -> str:
                return "an instance of {}".format(self.__class__.__name__)

        b = B()

        violation_error = None  # type: Optional[icontract.ViolationError]
        try:
            b.func()
        except icontract.ViolationError as err:
            violation_error = err

        self.assertIsNotNone(violation_error)
        self.assertEqual(
            textwrap.dedent(
                """\
                result % 2 == 0:
                result was 3
                self was an instance of B"""
            ),
            tests.error.wo_mandatory_location(str(violation_error)),
        )

    def test_ensure_then_violated_in_child(self) -> None:
        class A(icontract.DBC):
            @icontract.ensure(lambda result: result % 2 == 0)
            def func(self) -> int:
                return 10

        class B(A):
            @icontract.ensure(lambda result: result % 3 == 0)
            def func(self) -> int:
                return 2

            def __repr__(self) -> str:
                return "an instance of {}".format(self.__class__.__name__)

        b = B()

        violation_error = None  # type: Optional[icontract.ViolationError]
        try:
            b.func()
        except icontract.ViolationError as err:
            violation_error = err

        self.assertIsNotNone(violation_error)
        self.assertEqual(
            textwrap.dedent(
                """\
                result % 3 == 0:
                result was 2
                self was an instance of B"""
            ),
            tests.error.wo_mandatory_location(str(violation_error)),
        )

    def test_abstract_method(self) -> None:
        class A(icontract.DBC):
            @icontract.ensure(lambda result: result < 100)
            @abc.abstractmethod
            def func(self) -> int:
                pass

            def __repr__(self) -> str:
                return "Some A"

        class B(A):
            def func(self) -> int:
                return 1000

            def __repr__(self) -> str:
                return "an instance of {}".format(self.__class__.__name__)

        b = B()
        violation_error = None  # type: Optional[icontract.ViolationError]
        try:
            b.func()
        except icontract.ViolationError as err:
            violation_error = err

        self.assertIsNotNone(violation_error)
        self.assertEqual(
            textwrap.dedent(
                """\
                result < 100:
                result was 1000
                self was an instance of B"""
            ),
            tests.error.wo_mandatory_location(str(violation_error)),
        )

    def test_that_base_postconditions_apply_to_init_if_not_defined(self) -> None:
        class A(icontract.DBC):
            @icontract.ensure(lambda self: self.x >= 0)
            def __init__(self, x: int) -> None:
                self.x = x

            def __repr__(self) -> str:
                return "an instance of {}".format(self.__class__.__name__)

        class B(A):
            def __repr__(self) -> str:
                return "an instance of {}".format(self.__class__.__name__)

        violation_error = None  # type: Optional[icontract.ViolationError]
        try:
            _ = B(x=-1)
        except icontract.ViolationError as err:
            violation_error = err

        self.assertIsNotNone(violation_error)
        self.assertEqual(
            textwrap.dedent(
                """\
                self.x >= 0:
                result was None
                self was an instance of B
                self.x was -1
                x was -1"""
            ),
            tests.error.wo_mandatory_location(str(violation_error)),
        )

    def test_that_base_postconditions_dont_apply_to_init_if_overridden(self) -> None:
        class A(icontract.DBC):
            @icontract.ensure(lambda self: self.x >= 0)
            def __init__(self, x: int) -> None:
                self.x = x

        class B(A):
            # pylint: disable=super-init-not-called
            @icontract.ensure(lambda self: self.x < 0)
            def __init__(self, x: int) -> None:
                self.x = x

            def __repr__(self) -> str:
                return "an instance of {}".format(self.__class__.__name__)

        # postconditions of B need to be satisfied, but not from A
        _ = B(x=-100)

        violation_error = None  # Optional[icontract.ViolationError]
        try:
            _ = B(x=0)
        except icontract.ViolationError as err:
            violation_error = err

        self.assertIsNotNone(violation_error)
        self.assertEqual(
            textwrap.dedent(
                """\
                self.x < 0:
                result was None
                self was an instance of B
                self.x was 0
                x was 0"""
            ),
            tests.error.wo_mandatory_location(str(violation_error)),
        )


class TestPropertyOK(unittest.TestCase):
    def test_getter_setter_deleter_ok(self) -> None:
        class SomeBase(icontract.DBC):
            def __init__(self) -> None:
                self.deleted = False
                self._some_prop = 1

            @property
            @icontract.ensure(lambda self, result: self._some_prop == result)
            def some_prop(self) -> int:
                return self._some_prop

            @some_prop.setter
            @icontract.ensure(lambda self, value: self.some_prop == value)
            def some_prop(self, value: int) -> None:
                self._some_prop = value

            @some_prop.deleter
            @icontract.ensure(lambda self: self.deleted)
            def some_prop(self) -> None:
                self.deleted = True

        class SomeClass(SomeBase):
            pass

        some_inst = SomeClass()
        some_inst.some_prop = 3
        self.assertEqual(3, some_inst.some_prop)

        del some_inst.some_prop
        self.assertTrue(some_inst.deleted)


class TestPropertyViolation(unittest.TestCase):
    def test_getter(self) -> None:
        class SomeBase(icontract.DBC):
            def __init__(self) -> None:
                self.toggled = True

            @property
            @icontract.ensure(lambda self: not self.toggled)
            def some_prop(self) -> int:
                return 0

        class SomeClass(SomeBase):
            @property
            def some_prop(self) -> int:
                return 0

            def __repr__(self) -> str:
                return "an instance of {}".format(self.__class__.__name__)

        some_inst = SomeClass()

        violation_error = None  # type: Optional[icontract.ViolationError]
        try:
            _ = some_inst.some_prop
        except icontract.ViolationError as err:
            violation_error = err

        self.assertIsNotNone(violation_error)
        self.assertEqual(
            textwrap.dedent(
                """\
                not self.toggled:
                result was 0
                self was an instance of SomeClass
                self.toggled was True"""
            ),
            tests.error.wo_mandatory_location(str(violation_error)),
        )

    def test_setter(self) -> None:
        class SomeBase(icontract.DBC):
            def __init__(self) -> None:
                self.toggled = True

            @property
            def some_prop(self) -> int:
                return 0

            @some_prop.setter
            @icontract.ensure(lambda self: not self.toggled)
            def some_prop(self, value: int) -> None:
                pass

        class SomeClass(SomeBase):
            @SomeBase.some_prop.setter  # type: ignore
            def some_prop(self, value: int) -> None:
                pass

            def __repr__(self) -> str:
                return "an instance of {}".format(self.__class__.__name__)

        some_inst = SomeClass()

        violation_error = None  # type: Optional[icontract.ViolationError]
        try:
            some_inst.some_prop = 0
        except icontract.ViolationError as err:
            violation_error = err

        self.assertIsNotNone(violation_error)
        self.assertEqual(
            textwrap.dedent(
                """\
                not self.toggled:
                result was None
                self was an instance of SomeClass
                self.toggled was True
                value was 0"""
            ),
            tests.error.wo_mandatory_location(str(violation_error)),
        )

    def test_deleter(self) -> None:
        class SomeBase(icontract.DBC):
            def __init__(self) -> None:
                self.toggled = True

            @property
            def some_prop(self) -> int:
                return 0

            @some_prop.deleter
            @icontract.ensure(lambda self: not self.toggled)
            def some_prop(self) -> None:
                pass

        class SomeClass(SomeBase):
            def __repr__(self) -> str:
                return "an instance of {}".format(self.__class__.__name__)

            @SomeBase.some_prop.deleter  # type: ignore
            def some_prop(self) -> None:
                pass

        some_inst = SomeClass()

        violation_error = None  # type: Optional[icontract.ViolationError]
        try:
            del some_inst.some_prop
        except icontract.ViolationError as err:
            violation_error = err

        self.assertIsNotNone(violation_error)
        self.assertEqual(
            textwrap.dedent(
                """\
                not self.toggled:
                result was None
                self was an instance of SomeClass
                self.toggled was True"""
            ),
            tests.error.wo_mandatory_location(str(violation_error)),
        )

    def test_setter_strengthened(self) -> None:
        class SomeBase(icontract.DBC):
            def __init__(self) -> None:
                self.toggled = True

            @property
            def some_prop(self) -> int:
                return 0

            @some_prop.setter
            def some_prop(self, value: int) -> None:
                pass

        class SomeClass(SomeBase):
            @SomeBase.some_prop.setter  # type: ignore
            @icontract.ensure(lambda self: not self.toggled)
            def some_prop(self, value: int) -> None:
                pass

            def __repr__(self) -> str:
                return "an instance of {}".format(self.__class__.__name__)

        some_inst = SomeClass()

        violation_error = None  # type: Optional[icontract.ViolationError]
        try:
            some_inst.some_prop = 0
        except icontract.ViolationError as err:
            violation_error = err

        self.assertIsNotNone(violation_error)
        self.assertEqual(
            textwrap.dedent(
                """\
                not self.toggled:
                result was None
                self was an instance of SomeClass
                self.toggled was True
                value was 0"""
            ),
            tests.error.wo_mandatory_location(str(violation_error)),
        )


class TestInvalid(unittest.TestCase):
    def test_abstract_method_not_implemented(self) -> None:
        # pylint: disable=abstract-method
        class A(icontract.DBC):
            @icontract.ensure(lambda result: result < 100)
            @abc.abstractmethod
            def func(self) -> int:
                pass

        class B(A):
            pass

        type_err = None  # type: Optional[TypeError]
        try:
            _ = B()  # type: ignore
        except TypeError as err:
            type_err = err

        self.assertIsNotNone(type_err)
        if sys.version_info < (3, 9):
            self.assertEqual(
                "Can't instantiate abstract class B with abstract methods func",
                str(type_err),
            )
        elif sys.version_info < (3, 12):
            self.assertEqual(
                "Can't instantiate abstract class B with abstract method func",
                str(type_err),
            )
        else:
            self.assertEqual(
                "Can't instantiate abstract class B without an implementation for abstract method 'func'",
                str(type_err),
            )


if __name__ == "__main__":
    unittest.main()



#!/usr/bin/env python3
# pylint: disable=missing-docstring,invalid-name
# pylint: disable=unused-argument
import ast
import re
import textwrap
import unittest

import astor

# noinspection PyProtectedMember
import icontract._recompute


class TestTranslationForTracingAll(unittest.TestCase):
    @staticmethod
    def translate_all_expression(input_source_code: str) -> str:
        """
        Parse the input source code and translate it to a module with a tracing function.

        :param input_source_code: source code containing a single ``all`` expression
        :return:
            source code of the module with a tracing function.
            All non-deterministic bits are erased from it.
        """
        node = ast.parse(input_source_code)
        assert isinstance(node, ast.Module)
        expr_node = node.body[0]
        assert isinstance(expr_node, ast.Expr)
        call_node = expr_node.value
        assert isinstance(call_node, ast.Call)
        generator_exp = call_node.args[0]
        assert isinstance(generator_exp, ast.GeneratorExp)

        # We set only SOME_GLOBAL_CONSTANT in ``name_to_value``. In ``_recompute`` module
        # all the contract arguments will be set including also all the built-ins and other
        # global variables.

        module_node = icontract._recompute._translate_all_expression_to_a_module(
            generator_exp=generator_exp,
            generated_function_name="some_func",
            name_to_value={"SOME_GLOBAL_CONSTANT": 10},
        )

        got = astor.to_source(module_node)

        # We need to replace the UUID of the result variable for reproducibility.
        got = re.sub(
            r"icontract_tracing_all_result_[a-zA-Z0-9]+",
            "icontract_tracing_all_result",
            got,
        )

        assert isinstance(got, str)
        return got

    def test_global_variable(self) -> None:
        input_source_code = textwrap.dedent(
            """\
            all(
                x > SOME_GLOBAL_CONSTANT
                for x in lst
            )
            """
        )

        got_source_code = TestTranslationForTracingAll.translate_all_expression(
            input_source_code=input_source_code
        )

        # Please see ``TestTranslationForTracingAll.translate_all_expression`` and the note about ``name_to_value``
        # if you wonder why ``lst`` is not in the arguments.
        self.assertEqual(
            textwrap.dedent(
                """\
                def some_func(SOME_GLOBAL_CONSTANT):
                    for x in lst:
                        icontract_tracing_all_result = (x >
                            SOME_GLOBAL_CONSTANT)
                        if icontract_tracing_all_result:
                            pass
                        else:
                            return (
                                icontract_tracing_all_result,
                                (('x', x),))
                    return icontract_tracing_all_result, None
                """
            ),
            got_source_code,
        )

    def test_translation_two_fors_and_two_ifs(self) -> None:
        input_source_code = textwrap.dedent(
            """\
            all(
                cell > SOME_GLOBAL_CONSTANT
                for i, row in enumerate(matrix)
                if i > 0
                for j, cell in enumerate(row)
                if i == j
            )
            """
        )

        got_source_code = TestTranslationForTracingAll.translate_all_expression(
            input_source_code=input_source_code
        )

        # Please see ``TestTranslationForTracingAll.translate_all_expression`` and the note about ``name_to_value``
        # if you wonder why ``matrix`` is not in the arguments.
        self.assertEqual(
            textwrap.dedent(
                """\
                def some_func(SOME_GLOBAL_CONSTANT):
                    for i, row in enumerate(matrix):
                        if i > 0:
                            for j, cell in enumerate(row):
                                if i == j:
                                    (icontract_tracing_all_result
                                        ) = cell > SOME_GLOBAL_CONSTANT
                                    if (icontract_tracing_all_result
                                        ):
                                        pass
                                    else:
                                        return (
                                            icontract_tracing_all_result
                                            , (('i', i), ('row', row), ('j', j), ('cell',
                                            cell)))
                    return icontract_tracing_all_result, None
                """
            ),
            got_source_code,
        )

    def test_nested_all(self) -> None:
        # Nesting is not recursively followed by design. Only the outer-most all expression should be traced.

        input_source_code = textwrap.dedent(
            """\
            all(
                all(cell > SOME_GLOBAL_CONSTANT for cell in row)
                for row in matrix
            )
            """
        )

        got_source_code = TestTranslationForTracingAll.translate_all_expression(
            input_source_code=input_source_code
        )

        # Please see ``TestTranslationForTracingAll.translate_all_expression`` and the note about ``name_to_value``
        # if you wonder why ``matrix`` is not in the arguments.
        self.assertEqual(
            textwrap.dedent(
                """\
                def some_func(SOME_GLOBAL_CONSTANT):
                    for row in matrix:
                        icontract_tracing_all_result = all(
                            cell > SOME_GLOBAL_CONSTANT for cell in row)
                        if icontract_tracing_all_result:
                            pass
                        else:
                            return (
                                icontract_tracing_all_result,
                                (('row', row),))
                    return icontract_tracing_all_result, None
                """
            ),
            got_source_code,
        )


if __name__ == "__main__":
    unittest.main()



# pylint: disable=invalid-name
# pylint: disable=missing-docstring
# pylint: disable=unused-argument

import abc
import textwrap
import unittest
from typing import Optional  # pylint: disable=unused-import

import icontract
import tests.error


class TestOK(unittest.TestCase):
    def test_count_checks(self) -> None:
        class Increment:
            count = 0

            def __call__(self) -> bool:
                Increment.count += 1
                return True

        inc = Increment()

        @icontract.invariant(lambda self: inc())
        class A(icontract.DBC):
            def __repr__(self) -> str:
                return "instance of A"

            def some_func(self) -> int:
                return 1

        class B(A):
            def __repr__(self) -> str:
                return "instance of B"

            def some_func(self) -> int:
                return 2

        inst = B()
        self.assertEqual(
            1,
            Increment.count,
            "Invariant is expected to run only once at the initializer.",
        )

        inst.some_func()
        self.assertEqual(
            3,
            Increment.count,
            "Invariant is expected to run before and after the method call.",
        )

    def test_level_1_inheritance_of_invariants_does_not_leak_to_parents(self) -> None:
        # NOTE (mristin):
        # This is a regression test for:
        # https://github.com/Parquery/icontract/issues/295
        #
        # The invariants added to a child class were unexpectedly leaked back to
        # the parent class.

        @icontract.invariant(lambda: True)
        class Base(icontract.DBC):
            def do_something(self) -> None:
                pass

            def __repr__(self) -> str:
                return "instance of {}".format(self.__class__.__name__)

        @icontract.invariant(lambda: False)
        class Derived(Base):
            pass

        Base()

        # NOTE (mristin):
        # This produced an unexpected violation error.
        Base().do_something()

        had_violation_error = False
        try:
            Derived()
        except icontract.ViolationError:
            had_violation_error = True

        assert had_violation_error

    def test_level_2_inheritance_of_invariants_does_not_leak_to_parents(self) -> None:
        # NOTE (mristin):
        # This is a regression test for:
        # https://github.com/Parquery/icontract/issues/295
        #
        # The invariants added to a child class were unexpectedly leaked back to
        # the parent class.

        @icontract.invariant(lambda: True)
        class Base(icontract.DBC):
            def do_something(self) -> None:
                pass

            def __repr__(self) -> str:
                return "instance of {}".format(self.__class__.__name__)

        @icontract.invariant(lambda: True)
        class Derived(Base):
            pass

        @icontract.invariant(lambda: False)
        class DerivedDerived(Base):
            pass

        Base()
        Base().do_something()
        Derived()

        had_violation_error = False
        try:
            DerivedDerived()
        except icontract.ViolationError:
            had_violation_error = True

        assert had_violation_error

    # noinspection PyUnusedLocal
    def test_level_3_inheritance_of_invariants_does_not_leak_to_parents(self) -> None:
        # NOTE (mristin):
        # This is a regression test for:
        # https://github.com/Parquery/icontract/issues/295
        #
        # The invariants added to a child class were unexpectedly leaked back to
        # the parent class.

        class A(icontract.DBC):
            def do_something(self) -> None:
                pass

            def __repr__(self) -> str:
                return "instance of {}".format(self.__class__.__name__)

        @icontract.invariant(lambda: True)
        class B(A):
            pass

        # NOTE (mristin):
        # CFalse should not in any way influence A, B and CTrue, but it did due to
        # a bug.
        @icontract.invariant(lambda: False)
        class CFalse(B):  # pylint: disable=unused-variable
            pass

        @icontract.invariant(lambda: True)
        class CTrue(B):
            pass

        A()

        CTrue()

        A().do_something()

        # NOTE (mristin):
        # This produced an unexpected violation error.
        CTrue().do_something()

    def test_calling_super_init_checks_invariants_of_parent_not_child(self) -> None:
        # NOTE (mristin):
        # This test is a regression test from:
        # https://github.com/Parquery/icontract/issues/300

        @icontract.invariant(lambda self: True)
        class A(icontract.DBC):
            def __init__(self) -> None:
                pass

        @icontract.invariant(lambda self: self.x > 0)
        class B(A):
            def __init__(self, x: int) -> None:
                super().__init__()
                self.x = x

        _ = B(2)

    def test_calling_multi_super_init_checks_invariants_of_parent_not_child(
        self,
    ) -> None:
        # NOTE (mristin):
        # This test is a regression test from:
        # https://github.com/Parquery/icontract/issues/300

        @icontract.invariant(lambda self: True)
        class A1(icontract.DBC):
            def __init__(self) -> None:
                pass

        @icontract.invariant(lambda self: True)
        class A2(icontract.DBC):
            def __init__(self) -> None:
                pass

        @icontract.invariant(lambda self: self.x > 0)
        class B(A1, A2):
            def __init__(self, x: int) -> None:
                super().__init__()
                self.x = x

        _ = B(2)


class TestViolation(unittest.TestCase):
    def test_inherited(self) -> None:
        @icontract.invariant(lambda self: self.x > 0)
        class A(icontract.DBC):
            def __init__(self) -> None:
                self.x = 10

            def func(self) -> None:
                self.x = -1

            def __repr__(self) -> str:
                return "instance of A"

        class B(A):
            def __repr__(self) -> str:
                return "instance of B"

        b = B()
        violation_error = None  # type: Optional[icontract.ViolationError]
        try:
            b.func()
        except icontract.ViolationError as err:
            violation_error = err

        self.assertIsNotNone(violation_error)
        self.assertEqual(
            textwrap.dedent(
                """\
                self.x > 0:
                self was instance of B
                self.x was -1"""
            ),
            tests.error.wo_mandatory_location(str(violation_error)),
        )

    def test_inherited_violated_in_child(self) -> None:
        @icontract.invariant(lambda self: self.x > 0)
        class A(icontract.DBC):
            def __init__(self) -> None:
                self.x = 10

            def func(self) -> None:
                self.x = 100

            def __repr__(self) -> str:
                return "instance of A"

        class B(A):
            def func(self) -> None:
                self.x = -1

            def __repr__(self) -> str:
                return "instance of B"

        b = B()
        violation_error = None  # type: Optional[icontract.ViolationError]
        try:
            b.func()
        except icontract.ViolationError as err:
            violation_error = err

        self.assertIsNotNone(violation_error)
        self.assertEqual(
            textwrap.dedent(
                """\
                self.x > 0:
                self was instance of B
                self.x was -1"""
            ),
            tests.error.wo_mandatory_location(str(violation_error)),
        )

    def test_additional_invariant_violated_in_childs_init(self) -> None:
        @icontract.invariant(lambda self: self.x > 0)
        class A(icontract.DBC):
            def __init__(self) -> None:
                self.x = 10

            def __repr__(self) -> str:
                return "an instance of {}".format(self.__class__.__name__)

        @icontract.invariant(lambda self: self.x > 100)
        class B(A):
            pass

        violation_error = None  # type: Optional[icontract.ViolationError]
        try:
            _ = B()
        except icontract.ViolationError as err:
            violation_error = err

        self.assertIsNotNone(violation_error)
        self.assertEqual(
            textwrap.dedent(
                """\
                self.x > 100:
                self was an instance of B
                self.x was 10"""
            ),
            tests.error.wo_mandatory_location(str(violation_error)),
        )

    def test_method_violates_in_child(self) -> None:
        @icontract.invariant(lambda self: self.x > 0)
        class A(icontract.DBC):
            def __init__(self) -> None:
                self.x = 1000

            def some_method(self) -> None:
                self.x = 10

            def __repr__(self) -> str:
                return "an instance of A"

        @icontract.invariant(lambda self: self.x > 100)
        class B(A):
            def __repr__(self) -> str:
                return "an instance of B"

        b = B()
        violation_error = None  # type: Optional[icontract.ViolationError]
        try:
            b.some_method()
        except icontract.ViolationError as err:
            violation_error = err

        self.assertIsNotNone(violation_error)
        self.assertEqual(
            textwrap.dedent(
                """\
                self.x > 100:
                self was an instance of B
                self.x was 10"""
            ),
            tests.error.wo_mandatory_location(str(violation_error)),
        )

    def test_triple_inheritance(self) -> None:
        @icontract.invariant(lambda self: self.x > 0)
        class A(icontract.DBC):
            def __init__(self) -> None:
                self.x = 10

            def func(self) -> None:
                self.x = -1

            def __repr__(self) -> str:
                return "instance of A"

        class B(A):
            def __repr__(self) -> str:
                return "instance of B"

        class C(B):
            def __repr__(self) -> str:
                return "instance of C"

        c = C()
        violation_error = None  # type: Optional[icontract.ViolationError]
        try:
            c.func()
        except icontract.ViolationError as err:
            violation_error = err

        self.assertIsNotNone(violation_error)
        self.assertEqual(
            textwrap.dedent(
                """\
                self.x > 0:
                self was instance of C
                self.x was -1"""
            ),
            tests.error.wo_mandatory_location(str(violation_error)),
        )

    def test_with_abstract_method(self) -> None:
        @icontract.invariant(lambda self: self.x > 0)
        class A(icontract.DBC):
            def __init__(self) -> None:
                self.x = 10

            @abc.abstractmethod
            def func(self) -> None:
                pass

            def __repr__(self) -> str:
                return "an instance of A"

        class B(A):
            def func(self) -> None:
                self.x = -1

            def __repr__(self) -> str:
                return "an instance of B"

        b = B()
        violation_error = None  # type: Optional[icontract.ViolationError]
        try:
            b.func()
        except icontract.ViolationError as err:
            violation_error = err

        self.assertIsNotNone(violation_error)
        self.assertEqual(
            textwrap.dedent(
                """\
                self.x > 0:
                self was an instance of B
                self.x was -1"""
            ),
            tests.error.wo_mandatory_location(str(violation_error)),
        )

    def test_violated_in_super_init(self) -> None:
        # NOTE (mristin):
        # This test is a regression test from:
        # https://github.com/Parquery/icontract/issues/300

        @icontract.invariant(lambda self: self.x > 0)
        class A(icontract.DBC):
            def __init__(self, x: int):
                self.x = x

            def __repr__(self) -> str:
                return "an instance of {}".format(self.__class__.__name__)

        @icontract.invariant(lambda self: self.x > 10)
        class B(A):
            def __init__(self, x: int):  # pylint: disable=useless-parent-delegation
                super().__init__(x)

        violation_error = None  # type: Optional[icontract.ViolationError]
        try:
            _ = B(-1)
        except icontract.ViolationError as err:
            violation_error = err

        self.assertIsNotNone(violation_error)
        self.assertEqual(
            # NOTE (mristin):
            # This is an exception expected to be raised after super().__init__ call, not after B.__init__.
            textwrap.dedent(
                """\
                self.x > 0:
                self was an instance of B
                self.x was -1"""
            ),
            tests.error.wo_mandatory_location(str(violation_error)),
        )

    def test_violated_in_one_of_multi_super_init(self) -> None:
        # NOTE (mristin):
        # This test is a regression test from:
        # https://github.com/Parquery/icontract/issues/300

        @icontract.invariant(lambda self: self.x > 0)
        class A1(icontract.DBC):
            def __init__(self, x: int):
                self.x = x

            def __repr__(self) -> str:
                return "an instance of {}".format(self.__class__.__name__)

        @icontract.invariant(lambda self: self.x > 2)
        class A2(icontract.DBC):
            def __init__(self, x: int):
                self.x = x

        @icontract.invariant(lambda self: self.x > 10)
        class B(A1, A2):
            def __init__(self, x: int):
                A1.__init__(self, x)
                A2.__init__(self, x)

        violation_error = None  # type: Optional[icontract.ViolationError]
        try:
            # NOTE (mristin):
            # This is expected to violate __init__ in A2 but not A1.
            _ = B(1)
        except icontract.ViolationError as err:
            violation_error = err

        self.assertIsNotNone(violation_error)
        self.assertEqual(
            # NOTE (mristin):
            # ``self.x > 2`` indiactes a violation in __init__ of A2 but not A1 and not B.
            textwrap.dedent(
                """\
                self.x > 2:
                self was an instance of B
                self.x was 1"""
            ),
            tests.error.wo_mandatory_location(str(violation_error)),
        )

    def test_super_init_passes_but_init_violated(self) -> None:
        # NOTE (mristin):
        # This test is a regression test from:
        # https://github.com/Parquery/icontract/issues/300

        @icontract.invariant(lambda self: self.x > 0)
        class A(icontract.DBC):
            def __init__(self, x: int):
                self.x = x

            def __repr__(self) -> str:
                return "an instance of {}".format(self.__class__.__name__)

        @icontract.invariant(lambda self: self.x > 10)
        class B(A):
            def __init__(self, x: int):  # pylint: disable=useless-parent-delegation
                super().__init__(x)

        violation_error = None  # type: Optional[icontract.ViolationError]
        try:
            _ = B(2)
        except icontract.ViolationError as err:
            violation_error = err

        self.assertIsNotNone(violation_error)
        self.assertEqual(
            # NOTE (mristin):
            # This is an exception expected to be raised after B.__init__, but not earlier after super().__init__ call.
            textwrap.dedent(
                """\
                self.x > 10:
                self was an instance of B
                self.x was 2"""
            ),
            tests.error.wo_mandatory_location(str(violation_error)),
        )

    def test_multi_super_inits_pass_but_init_violated(self) -> None:
        # NOTE (mristin):
        # This test is a regression test from:
        # https://github.com/Parquery/icontract/issues/300

        @icontract.invariant(lambda self: self.x > 0)
        class A1(icontract.DBC):
            def __init__(self, x: int):
                self.x = x

            def __repr__(self) -> str:
                return "an instance of {}".format(self.__class__.__name__)

        @icontract.invariant(lambda self: self.x > 2)
        class A2(icontract.DBC):
            def __init__(self, x: int):
                self.x = x

            def __repr__(self) -> str:
                return "an instance of {}".format(self.__class__.__name__)

        @icontract.invariant(lambda self: self.x > 10)
        class B(A1, A2):
            def __init__(self, x: int):
                A1.__init__(self, x)
                A2.__init__(self, x)

        violation_error = None  # type: Optional[icontract.ViolationError]
        try:
            _ = B(3)
        except icontract.ViolationError as err:
            violation_error = err

        self.assertIsNotNone(violation_error)
        self.assertEqual(
            # NOTE (mristin):
            # This is an exception expected to be raised after B.__init__, but not in any super __init__ calls.
            textwrap.dedent(
                """\
                self.x > 10:
                self was an instance of B
                self.x was 3"""
            ),
            tests.error.wo_mandatory_location(str(violation_error)),
        )

    def test_invariant_violated_in_parent_in_implicit_init(self) -> None:
        @icontract.invariant(lambda self: self.x > 0)
        class A(icontract.DBC):
            def __init__(self, x: int) -> None:
                self.x = x

            def __repr__(self) -> str:
                return "an instance of {}".format(self.__class__.__name__)

        @icontract.invariant(lambda self: self.x > 10)
        class B(A):
            # NOTE (mristin):
            # We explicitly define no constructor (``__init__``) here. The constructor is implicitly called
            # from A (``A.__init__``). However, after the call to the implicit constructor, the invariants of B,
            # and not only A, should apply.
            pass

        # NOTE (mristin):
        # We check that the invariants of the parent are violated here.
        violation_error = None  # type: Optional[icontract.ViolationError]
        try:
            _ = B(-1)
        except icontract.ViolationError as err:
            violation_error = err

        self.assertIsNotNone(violation_error)
        self.assertEqual(
            # NOTE (mristin):
            # This is a violation expected to be raised after A.__init__ which is delegated to from B.__init__.
            textwrap.dedent(
                """\
                self.x > 0:
                self was an instance of B
                self.x was -1"""
            ),
            tests.error.wo_mandatory_location(str(violation_error)),
        )

    def test_invariant_violated_in_one_of_parents_in_implicit_init(self) -> None:
        @icontract.invariant(lambda self: self.x > 0)
        class A1(icontract.DBC):
            def __init__(self, x: int) -> None:
                self.x = x

            def __repr__(self) -> str:
                return "an instance of {}".format(self.__class__.__name__)

        @icontract.invariant(lambda self: self.x > 5)
        class A2(icontract.DBC):
            def __init__(self, x: int) -> None:
                self.x = x

        class B(A1, A2):
            # NOTE (mristin):
            # We explicitly define no constructor (``__init__``) here. The constructor is implicitly called
            # from one of the parents (``A1.__init__`` or ``A2.__init__``). However, after the call to the implicit
            # constructor, the invariants of B, and not only of A1 and A2, should apply.
            pass

        # NOTE (mristin):
        # We check that the invariants of one of the parents are violated here.
        violation_error = None  # type: Optional[icontract.ViolationError]
        try:
            _ = B(3)
        except icontract.ViolationError as err:
            violation_error = err

        self.assertIsNotNone(violation_error)
        self.assertEqual(
            # NOTE (mristin):
            # This is a violation expected to be raised after A2.__init__ which is delegated to from B.__init__.
            textwrap.dedent(
                """\
                self.x > 5:
                self was an instance of B
                self.x was 3"""
            ),
            tests.error.wo_mandatory_location(str(violation_error)),
        )

    def test_invariant_violated_in_child_in_implicit_init(self) -> None:
        @icontract.invariant(lambda self: self.x > 0)
        class A(icontract.DBC):
            def __init__(self, x: int) -> None:
                self.x = x

            def __repr__(self) -> str:
                return "an instance of {}".format(self.__class__.__name__)

        @icontract.invariant(lambda self: self.x > 10)
        class B(A):
            # NOTE (mristin):
            # We explicitly define no constructor (``__init__``) here. The constructor is implicitly called
            # from A (``A.__init__``). However, after the call to the implicit constructor, the invariants of B,
            # and not only A, should apply.
            pass

        # NOTE (mristin):
        # We make sure explicitly that the invariants did not leak to the parent -- the invariants of B should not
        # apply to an instance of A.
        _ = A(3)

        # NOTE (mristin):
        # We also double-check that the invariants pass correctly.
        _ = B(11)

        # NOTE (mristin):
        # Now we check that the invariants are violated which we exactly expect.
        violation_error = None  # type: Optional[icontract.ViolationError]
        try:
            _ = B(3)
        except icontract.ViolationError as err:
            violation_error = err

        self.assertIsNotNone(violation_error)
        self.assertEqual(
            # NOTE (mristin):
            # This is a violation expected to be raised after B.__init__ even though B.__init__ is simply delegated to
            # A.__init__.
            textwrap.dedent(
                """\
                self.x > 10:
                self was an instance of B
                self.x was 3"""
            ),
            tests.error.wo_mandatory_location(str(violation_error)),
        )

    def test_invariant_violated_in_child_in_implicit_init_with_multiple_inheritance(
        self,
    ) -> None:
        @icontract.invariant(lambda self: self.x > 0)
        class A1(icontract.DBC):
            def __init__(self, x: int) -> None:
                self.x = x

            def __repr__(self) -> str:
                return "an instance of {}".format(self.__class__.__name__)

        @icontract.invariant(lambda self: self.x > 5)
        class A2(icontract.DBC):
            def __init__(self, x: int) -> None:
                self.x = x

        @icontract.invariant(lambda self: self.x > 10)
        class B(A1, A2):
            # NOTE (mristin):
            # We explicitly define no constructor (``__init__``) here. The constructor is implicitly called
            # from one of the super classes. However, after the call to the implicit constructor, the invariants of B,
            # and not only of A1 and A2, should apply.
            pass

        # NOTE (mristin):
        # We make sure explicitly that the invariants did not leak to the parent -- the invariants of B should not
        # apply to an instance of A1 or of A2.
        _ = A1(3)
        _ = A2(7)

        # NOTE (mristin):
        # We also double-check that the invariant checks can pass correctly.
        _ = B(11)

        # NOTE (mristin):
        # Now we check that the invariants are violated which we exactly expect.
        violation_error = None  # type: Optional[icontract.ViolationError]
        try:
            _ = B(8)
        except icontract.ViolationError as err:
            violation_error = err

        self.assertIsNotNone(violation_error)
        self.assertEqual(
            # NOTE (mristin):
            # This is a violation expected to be raised after B.__init__ even though B.__init__ is simply delegated to
            # either A1.__init__ or A2.__init__.
            textwrap.dedent(
                """\
                self.x > 10:
                self was an instance of B
                self.x was 8"""
            ),
            tests.error.wo_mandatory_location(str(violation_error)),
        )


class TestProperty(unittest.TestCase):
    def test_inherited_getter(self) -> None:
        @icontract.invariant(lambda self: not self.toggled)
        class SomeBase(icontract.DBC):
            def __init__(self) -> None:
                self.toggled = False

            @property
            def some_prop(self) -> int:
                self.toggled = True
                return 0

        class SomeClass(SomeBase):
            def __repr__(self) -> str:
                return "an instance of {}".format(self.__class__.__name__)

        some_inst = SomeClass()

        violation_error = None  # type: Optional[icontract.ViolationError]
        try:
            _ = some_inst.some_prop
        except icontract.ViolationError as err:
            violation_error = err

        self.assertIsNotNone(violation_error)
        self.assertEqual(
            textwrap.dedent(
                """\
                not self.toggled:
                self was an instance of SomeClass
                self.toggled was True"""
            ),
            tests.error.wo_mandatory_location(str(violation_error)),
        )

    def test_inherited_setter(self) -> None:
        @icontract.invariant(lambda self: not self.toggled)
        class SomeBase(icontract.DBC):
            def __init__(self) -> None:
                self.toggled = False

            @property
            def some_prop(self) -> int:
                return 0

            @some_prop.setter
            def some_prop(self, value: int) -> None:
                self.toggled = True

        class SomeClass(SomeBase):
            def __repr__(self) -> str:
                return "an instance of {}".format(self.__class__.__name__)

        some_inst = SomeClass()

        violation_error = None  # type: Optional[icontract.ViolationError]
        try:
            some_inst.some_prop = 0
        except icontract.ViolationError as err:
            violation_error = err

        self.assertIsNotNone(violation_error)
        self.assertEqual(
            textwrap.dedent(
                """\
                not self.toggled:
                self was an instance of SomeClass
                self.toggled was True"""
            ),
            tests.error.wo_mandatory_location(str(violation_error)),
        )

    def test_inherited_deleter(self) -> None:
        @icontract.invariant(lambda self: not self.toggled)
        class SomeBase(icontract.DBC):
            def __init__(self) -> None:
                self.toggled = False

            @property
            def some_prop(self) -> int:
                return 0

            @some_prop.deleter
            def some_prop(self) -> None:
                self.toggled = True

        class SomeClass(SomeBase):
            def __repr__(self) -> str:
                return "an instance of {}".format(self.__class__.__name__)

        some_inst = SomeClass()

        violation_error = None  # type: Optional[icontract.ViolationError]
        try:
            del some_inst.some_prop
        except icontract.ViolationError as err:
            violation_error = err

        self.assertIsNotNone(violation_error)
        self.assertEqual(
            textwrap.dedent(
                """\
                not self.toggled:
                self was an instance of SomeClass
                self.toggled was True"""
            ),
            tests.error.wo_mandatory_location(str(violation_error)),
        )

    def test_inherited_invariant_on_getter(self) -> None:
        @icontract.invariant(lambda self: not self.toggled)
        class SomeBase(icontract.DBC):
            def __init__(self) -> None:
                self.toggled = False

        class SomeClass(SomeBase):
            @property
            def some_prop(self) -> int:
                self.toggled = True
                return 0

            def __repr__(self) -> str:
                return "an instance of {}".format(self.__class__.__name__)

        some_inst = SomeClass()

        violation_error = None  # type: Optional[icontract.ViolationError]
        try:
            _ = some_inst.some_prop
        except icontract.ViolationError as err:
            violation_error = err

        self.assertIsNotNone(violation_error)
        self.assertEqual(
            textwrap.dedent(
                """\
                not self.toggled:
                self was an instance of SomeClass
                self.toggled was True"""
            ),
            tests.error.wo_mandatory_location(str(violation_error)),
        )

    def test_inherited_invariant_on_setter(self) -> None:
        @icontract.invariant(lambda self: not self.toggled)
        class SomeBase(icontract.DBC):
            def __init__(self) -> None:
                self.toggled = False

        class SomeClass(SomeBase):
            @property
            def some_prop(self) -> int:
                return 0

            @some_prop.setter
            def some_prop(self, value: int) -> None:
                self.toggled = True

            def __repr__(self) -> str:
                return "an instance of {}".format(self.__class__.__name__)

        some_inst = SomeClass()

        violation_error = None  # type: Optional[icontract.ViolationError]
        try:
            some_inst.some_prop = 0
        except icontract.ViolationError as err:
            violation_error = err

        self.assertIsNotNone(violation_error)
        self.assertEqual(
            textwrap.dedent(
                """\
                not self.toggled:
                self was an instance of SomeClass
                self.toggled was True"""
            ),
            tests.error.wo_mandatory_location(str(violation_error)),
        )

    def test_inherited_invariant_on_deleter(self) -> None:
        @icontract.invariant(lambda self: not self.toggled)
        class SomeBase(icontract.DBC):
            def __init__(self) -> None:
                self.toggled = False

        class SomeClass(SomeBase):
            @property
            def some_prop(self) -> int:
                return 0

            @some_prop.deleter
            def some_prop(self) -> None:
                self.toggled = True

            def __repr__(self) -> str:
                return "an instance of {}".format(self.__class__.__name__)

        some_inst = SomeClass()

        violation_error = None  # type: Optional[icontract.ViolationError]
        try:
            del some_inst.some_prop
        except icontract.ViolationError as err:
            violation_error = err

        self.assertIsNotNone(violation_error)
        self.assertEqual(
            textwrap.dedent(
                """\
                not self.toggled:
                self was an instance of SomeClass
                self.toggled was True"""
            ),
            tests.error.wo_mandatory_location(str(violation_error)),
        )


if __name__ == "__main__":
    unittest.main()



# pylint: disable=missing-docstring
import pathlib
import subprocess
import sys
import tempfile
import unittest

if sys.version_info < (3, 8):
    raise unittest.SkipTest(
        (
            "Mypy since 1.5 dropped support for Python 3.7 and "
            "you are running Python {}, so skipping."
        ).format(sys.version_info)
    )


class TestMypyDecorators(unittest.TestCase):
    def test_functions(self) -> None:
        with tempfile.TemporaryDirectory(prefix="mypy_fail_case_") as tmpdir:
            content = '''\
"""Implement a fail case for mypy to test that the types are preserved with the decorators."""

import icontract

@icontract.require(lambda x: x > 0)
def f1(x: int) -> int:
    return x
f1("this is wrong")

@icontract.ensure(lambda result: result > 0)
def f2(x: int) -> int:
    return x
f2("this is wrong")

@icontract.snapshot(lambda x: x)
def f3(x: int) -> int:
    return x
f3("this is wrong")
'''

            pth = pathlib.Path(tmpdir) / "source.py"
            pth.write_text(content, encoding="utf-8")

            with subprocess.Popen(
                ["mypy", "--strict", str(pth)],
                universal_newlines=True,
                stdout=subprocess.PIPE,
            ) as proc:
                out, err = proc.communicate()

                self.assertIsNone(err)
                self.assertEqual(
                    """\
{path}:8: error: Argument 1 to "f1" has incompatible type "str"; expected "int"  [arg-type]
{path}:13: error: Argument 1 to "f2" has incompatible type "str"; expected "int"  [arg-type]
{path}:18: error: Argument 1 to "f3" has incompatible type "str"; expected "int"  [arg-type]
Found 3 errors in 1 file (checked 1 source file)
""".format(
                        path=pth
                    ),
                    out,
                )

    def test_class_type_when_decorated_with_invariant(self) -> None:
        with tempfile.TemporaryDirectory(prefix="mypy_fail_case_") as tmpdir:
            content = '''\
"""Implement a passing case for mypy to test that the type of class is preserved."""

import icontract

class SomeClass:
    pass

reveal_type(SomeClass)
    
@icontract.invariant(lambda self: self.x > 0)
class Decorated:
    def __init__(self) -> None:
        self.x = 1

reveal_type(Decorated)
'''

            pth = pathlib.Path(tmpdir) / "source.py"
            pth.write_text(content, encoding="utf-8")

            with subprocess.Popen(
                ["mypy", "--strict", str(pth)],
                universal_newlines=True,
                stdout=subprocess.PIPE,
            ) as proc:
                out, err = proc.communicate()

                self.assertIsNone(err)
                self.assertEqual(
                    """\
{path}:8: note: Revealed type is "def () -> source.SomeClass"
{path}:15: note: Revealed type is "def () -> source.Decorated"
Success: no issues found in 1 source file
""".format(
                        path=pth
                    ),
                    out,
                )

    def test_that_mypy_complains_when_decorating_non_type_with_invariant(self) -> None:
        with tempfile.TemporaryDirectory(prefix="mypy_fail_case_") as tmpdir:
            content = '''\
"""Provide a fail case to test that mypy complains when we decorate a non-type with invariant."""

import icontract

@icontract.invariant(lambda: True)
def some_func() -> None:
    pass
'''

            pth = pathlib.Path(tmpdir) / "source.py"
            pth.write_text(content, encoding="utf-8")

            with subprocess.Popen(
                ["mypy", "--strict", str(pth)],
                universal_newlines=True,
                stdout=subprocess.PIPE,
            ) as proc:
                out, err = proc.communicate()

                self.assertIsNone(err)
                self.assertEqual(
                    """\
{path}:5: error: Value of type variable "ClassT" of "__call__" of "invariant" cannot be "Callable[[], None]"  [type-var]
Found 1 error in 1 file (checked 1 source file)
""".format(
                        path=pth
                    ),
                    out,
                )


if __name__ == "__main__":
    unittest.main()



# pylint: disable=missing-docstring
# pylint: disable=invalid-name
# pylint: disable=unused-argument
# pylint: disable=unused-variable

import unittest
from typing import Optional

import icontract

import tests.error


class TestNoneSpecified(unittest.TestCase):
    def test_that_it_works(self) -> None:
        @icontract.require(lambda x: x > 0)
        def some_func(x: int) -> None:
            pass

        violation_error = None  # type: Optional[icontract.ViolationError]
        try:
            some_func(x=-1)
        except icontract.ViolationError as err:
            violation_error = err

        assert violation_error is not None
        self.assertEqual(
            "x > 0: x was -1", tests.error.wo_mandatory_location(str(violation_error))
        )


class TestSpecifiedAsFunction(unittest.TestCase):
    def test_lambda(self) -> None:
        @icontract.require(
            lambda x: x > 0,
            error=lambda x: ValueError("x must be positive: {}".format(x)),
        )
        def some_func(x: int) -> None:
            pass

        value_error = None  # type: Optional[ValueError]
        try:
            some_func(x=-1)
        except ValueError as err:
            value_error = err

        assert value_error is not None
        self.assertEqual("x must be positive: -1", str(value_error))

    def test_separate_function(self) -> None:
        def error_func(x: int) -> ValueError:
            return ValueError("x must be positive: {}".format(x))

        @icontract.require(lambda x: x > 0, error=error_func)
        def some_func(x: int) -> None:
            pass

        value_error = None  # type: Optional[ValueError]
        try:
            some_func(x=-1)
        except ValueError as err:
            value_error = err

        assert value_error is not None
        self.assertEqual("x must be positive: -1", str(value_error))

    def test_separate_method(self) -> None:
        class Errorer:
            def error_func(self, x: int) -> ValueError:
                return ValueError("x must be positive: {}".format(x))

        errorer = Errorer()

        @icontract.require(lambda x: x > 0, error=errorer.error_func)
        def some_func(x: int) -> None:
            pass

        value_error = None  # type: Optional[ValueError]
        try:
            some_func(x=-1)
        except ValueError as err:
            value_error = err

        assert value_error is not None
        self.assertEqual("x must be positive: -1", str(value_error))

    def test_report_if_result_is_not_base_exception(self) -> None:
        @icontract.require(lambda x: x > 0, error=lambda x: "x must be positive")  # type: ignore
        def some_func(x: int) -> None:
            pass

        type_error = None  # type: Optional[TypeError]
        try:
            some_func(x=-1)
        except TypeError as err:
            type_error = err

        assert type_error is not None
        self.assertRegex(
            str(type_error),
            r"^The exception returned by the contract's error <function .*> does not inherit from BaseException\.$",
        )


class TestSpecifiedAsType(unittest.TestCase):
    def test_valid_exception(self) -> None:
        @icontract.require(lambda x: x > 0, error=ValueError)
        def some_func(x: int) -> None:
            pass

        value_error = None  # type: Optional[ValueError]
        try:
            some_func(x=-1)
        except ValueError as err:
            value_error = err

        assert value_error is not None
        self.assertEqual(
            "x > 0: x was -1", tests.error.wo_mandatory_location(str(value_error))
        )


class TestSpecifiedAsInstance(unittest.TestCase):
    def test_valid_exception(self) -> None:
        @icontract.require(lambda x: x > 0, error=ValueError("negative x"))
        def some_func(x: int) -> None:
            pass

        value_error = None  # type: Optional[ValueError]
        try:
            some_func(x=-1)
        except ValueError as err:
            value_error = err

        assert value_error is not None
        self.assertEqual("negative x", str(value_error))

    def test_repeated_raising(self) -> None:
        @icontract.require(lambda x: x > 0, error=ValueError("negative x"))
        def some_func(x: int) -> None:
            pass

        value_error = None  # type: Optional[ValueError]
        try:
            some_func(x=-1)
        except ValueError as err:
            value_error = err

        assert value_error is not None
        self.assertEqual("negative x", str(value_error))

        # Repeat
        value_error = None
        try:
            some_func(x=-1)
        except ValueError as err:
            value_error = err

        assert value_error is not None
        self.assertEqual("negative x", str(value_error))


class TestSpecifiedAsInvalidType(unittest.TestCase):
    def test_in_precondition(self) -> None:
        class A:
            pass

        value_error = None  # type: Optional[ValueError]
        try:

            @icontract.require(lambda x: x > 0, error=A)  # type: ignore
            def some_func(x: int) -> None:
                pass

        except ValueError as err:
            value_error = err

        assert value_error is not None
        self.assertRegex(
            str(value_error),
            r"The error of the contract is given as a type, "
            r"but the type does not inherit from BaseException: <class .*\.A'>",
        )

    def test_in_postcondition(self) -> None:
        class A:
            pass

        value_error = None  # type: Optional[ValueError]
        try:

            @icontract.ensure(lambda result: result > 0, error=A)  # type: ignore
            def some_func() -> int:
                return -1

        except ValueError as err:
            value_error = err

        assert value_error is not None
        self.assertRegex(
            str(value_error),
            r"The error of the contract is given as a type, "
            r"but the type does not inherit from BaseException: <class .*\.A'>",
        )

    def test_in_invariant(self) -> None:
        value_error = None  # type: Optional[ValueError]
        try:

            class A:
                pass

            @icontract.invariant(lambda self: self.x > 0, error=A)  # type: ignore
            class B:
                def __init__(self) -> None:
                    self.x = -1

        except ValueError as err:
            value_error = err

        assert value_error is not None
        self.assertRegex(
            str(value_error),
            r"The error of the contract is given as a type, "
            r"but the type does not inherit from BaseException: <class .*\.A'>",
        )


class TestSpecifiedAsInstanceOfInvalidType(unittest.TestCase):
    def test_in_precondition(self) -> None:
        class A:
            def __init__(self, msg: str) -> None:
                self.msg = msg

        value_error = None  # type: Optional[ValueError]
        try:

            @icontract.require(lambda x: x > 0, error=A("something went wrong"))  # type: ignore
            def some_func(x: int) -> None:
                pass

        except ValueError as err:
            value_error = err

        assert value_error is not None
        self.assertRegex(
            str(value_error),
            r"^The error of the contract must be either a callable \(a function or a method\), "
            r"a class \(subclass of BaseException\) or an instance of BaseException, "
            r"but got: <.*\.A object at 0x.*>$",
        )

    def test_in_postcondition(self) -> None:
        class A:
            def __init__(self, msg: str) -> None:
                self.msg = msg

        value_error = None  # type: Optional[ValueError]
        try:

            @icontract.ensure(lambda result: result > 0, error=A("something went wrong"))  # type: ignore
            def some_func() -> int:
                return -1

        except ValueError as err:
            value_error = err

        assert value_error is not None
        self.assertRegex(
            str(value_error),
            r"^The error of the contract must be either a callable \(a function or a method\), "
            r"a class \(subclass of BaseException\) or an instance of BaseException, "
            r"but got: <.*\.A object at 0x.*>$",
        )

    def test_in_invariant(self) -> None:
        class A:
            def __init__(self, msg: str) -> None:
                self.msg = msg

        value_error = None  # type: Optional[ValueError]
        try:

            @icontract.invariant(lambda self: self.x > 0, error=A("something went wrong"))  # type: ignore
            class B:
                def __init__(self) -> None:
                    self.x = -1

        except ValueError as err:
            value_error = err

        assert value_error is not None
        self.assertRegex(
            str(value_error),
            r"^The error of the contract must be either a callable \(a function or a method\), "
            r"a class \(subclass of BaseException\) or an instance of BaseException, "
            r"but got: <.*\.A object at 0x.*>$",
        )


if __name__ == "__main__":
    unittest.main()



# pylint: disable=missing-docstring
# pylint: disable=invalid-name
# pylint: disable=unused-argument

import textwrap
import unittest
from typing import Optional

import typeguard

import icontract
import tests.error


class TestPrecondition(unittest.TestCase):
    def test_both_precondition_and_typeguard_ok(self) -> None:
        @icontract.require(lambda x: x > 0)
        @typeguard.typechecked
        def some_func(x: int) -> None:
            pass

        some_func(1)

    def test_precondition_ok_and_typeguard_fails(self) -> None:
        class A:
            def is_ok(self) -> bool:
                return True

        class B:
            def is_ok(self) -> bool:
                return True

        @icontract.require(lambda x: x.is_ok())
        @typeguard.typechecked
        def some_func(x: A) -> None:
            pass

        b = B()
        type_check_error = None  # type: Optional[typeguard.TypeCheckError]
        try:
            some_func(b)  # type: ignore
        except typeguard.TypeCheckError as err:
            type_check_error = err

        self.assertIsNotNone(type_check_error)

    def test_precondition_fails_and_typeguard_ok(self) -> None:
        @icontract.require(lambda x: x > 0)
        @typeguard.typechecked
        def some_func(x: int) -> None:
            pass

        violation_error = None  # type: Optional[icontract.ViolationError]
        try:
            some_func(-10)
        except icontract.ViolationError as err:
            violation_error = err

        self.assertIsNotNone(violation_error)
        self.assertEqual(
            "x > 0: x was -10", tests.error.wo_mandatory_location(str(violation_error))
        )


class TestInvariant(unittest.TestCase):
    def test_both_invariant_and_typeguard_ok(self) -> None:
        @icontract.invariant(lambda self: self.x > 0)
        @typeguard.typechecked
        class A:
            def __init__(self, x: int) -> None:
                self.x = x

            def do_something(self, y: int) -> None:
                self.x += y

        a = A(x=1)
        a.do_something(y=2)

    def test_invariant_ok_and_typeguard_fails(self) -> None:
        class A:
            def is_ok(self) -> bool:
                return True

        class B:
            def is_ok(self) -> bool:
                return True

        @icontract.invariant(lambda self: self.x.is_ok())
        @typeguard.typechecked
        class C:
            def __init__(self, a: A) -> None:
                self.a = a

        b = B()

        type_check_error = None  # type: Optional[typeguard.TypeCheckError]
        try:
            _ = C(a=b)  # type: ignore
        except typeguard.TypeCheckError as err:
            type_check_error = err

        self.assertIsNotNone(type_check_error)

    def test_invariant_fails_and_typeguard_ok(self) -> None:
        @icontract.invariant(lambda self: self.x > 0)
        @typeguard.typechecked
        class A:
            def __init__(self, x: int) -> None:
                self.x = x

            def __repr__(self) -> str:
                return "an instance of {}".format(self.__class__.__name__)

        violation_error = None  # type: Optional[icontract.ViolationError]
        try:
            _ = A(-1)
        except icontract.ViolationError as err:
            violation_error = err

        self.assertIsNotNone(violation_error)
        self.assertEqual(
            textwrap.dedent(
                """\
                self.x > 0:
                self was an instance of A
                self.x was -1"""
            ),
            tests.error.wo_mandatory_location(str(violation_error)),
        )


class TestInheritance(unittest.TestCase):
    def test_both_invariant_and_typeguard_ok(self) -> None:
        @icontract.invariant(lambda self: self.x > 0)
        @typeguard.typechecked
        class A(icontract.DBC):
            def __init__(self, x: int) -> None:
                self.x = x

            def do_something(self, y: int) -> None:
                self.x += y

        class B(A):
            pass

        b = B(x=1)
        b.do_something(y=2)

    def test_invariant_ok_and_typeguard_fails(self) -> None:
        class A:
            def is_ok(self) -> bool:
                return True

        class B:
            def is_ok(self) -> bool:
                return True

        @icontract.invariant(lambda self: self.x.is_ok())
        @typeguard.typechecked
        class C(icontract.DBC):
            def __init__(self, a: A) -> None:
                self.a = a

        class D(C):
            pass

        b = B()

        type_check_error = None  # type: Optional[typeguard.TypeCheckError]
        try:
            _ = D(a=b)  # type: ignore
        except typeguard.TypeCheckError as err:
            type_check_error = err

        self.assertIsNotNone(type_check_error)

    def test_invariant_fails_and_typeguard_ok(self) -> None:
        @icontract.invariant(lambda self: self.x > 0)
        @typeguard.typechecked
        class A:
            def __init__(self, x: int) -> None:
                self.x = x

            def __repr__(self) -> str:
                return "an instance of {}".format(self.__class__.__name__)

        class B(A):
            def __repr__(self) -> str:
                return "an instance of {}".format(self.__class__.__name__)

        violation_error = None  # type: Optional[icontract.ViolationError]
        try:
            _ = B(-1)
        except icontract.ViolationError as err:
            violation_error = err

        self.assertIsNotNone(violation_error)
        self.assertEqual(
            textwrap.dedent(
                """\
                self.x > 0:
                self was an instance of B
                self.x was -1"""
            ),
            tests.error.wo_mandatory_location(str(violation_error)),
        )


if __name__ == "__main__":
    unittest.main()



"""Test logic that can be potentially used by the integrators such as third-party libraries."""

# pylint: disable=missing-docstring
# pylint: disable=invalid-name,unnecessary-lambda

import ast
import unittest
from typing import List, MutableMapping, Any, Optional

import icontract._checkers
import icontract._represent


@icontract.require(lambda x: x > 0)
@icontract.snapshot(
    lambda cumulative: None if len(cumulative) == 0 else cumulative[-1], "last"
)
@icontract.snapshot(lambda cumulative: len(cumulative), "len_cumulative")
@icontract.ensure(lambda cumulative, OLD: len(cumulative) == OLD.len_cumulative + 1)
@icontract.ensure(
    lambda x, cumulative, OLD: OLD.last is None or OLD.last + x == cumulative[-1]
)
@icontract.ensure(
    lambda x, cumulative, OLD: OLD.last is not None or x == cumulative[-1]
)
def func_with_contracts(x: int, cumulative: List[int]) -> None:
    if len(cumulative) == 0:
        cumulative.append(x)
    else:
        cumulative.append(x + cumulative[-1])


def func_without_contracts() -> None:
    pass


@icontract.invariant(lambda self: self.x > 0)
class ClassWithInvariants(icontract.DBC):
    def __init__(self) -> None:
        self.x = 1


class TestInitial(unittest.TestCase):
    def test_that_there_is_no_checker_if_no_contracts(self) -> None:
        checker = icontract._checkers.find_checker(func=func_without_contracts)
        self.assertIsNone(checker)


class TestPreconditions(unittest.TestCase):
    def test_evaluating(self) -> None:
        checker = icontract._checkers.find_checker(func=func_with_contracts)
        assert checker is not None

        preconditions = checker.__preconditions__  # type: ignore
        assert isinstance(preconditions, list)
        assert all(isinstance(group, list) for group in preconditions)
        assert all(
            isinstance(contract, icontract._types.Contract)
            for group in preconditions
            for contract in group
        )

        ##
        # Evaluate manually preconditions
        ##

        kwargs = {"x": 4, "cumulative": [2]}

        success = True
        # We have to check preconditions in groups in case they are weakened
        for group in preconditions:
            success = True
            for contract in group:
                condition_kwargs = icontract._checkers.select_condition_kwargs(
                    contract=contract, resolved_kwargs=kwargs
                )

                success = contract.condition(**condition_kwargs)
                if not success:
                    break

            if success:
                break

        assert success

    def test_adding(self) -> None:
        def some_func(x: int) -> None:  # pylint: disable=unused-argument
            return

        checker = icontract._checkers.find_checker(func=some_func)
        assert checker is None

        wrapped = checker = icontract._checkers.decorate_with_checker(func=some_func)

        # The contract needs to have its own error specified since it is not added as a decorator,
        # so the module ``icontract._represent`` will be confused.
        icontract._checkers.add_precondition_to_checker(
            checker=checker,
            contract=icontract._types.Contract(
                condition=lambda x: x > 0,
                error=lambda x: icontract.ViolationError(
                    "x must be positive, but got: {}".format(x)
                ),
            ),
        )

        violation_error = None  # type: Optional[icontract.ViolationError]
        try:
            wrapped(x=-1)
        except icontract.ViolationError as err:
            violation_error = err

        assert violation_error is not None

        self.assertEqual("x must be positive, but got: -1", str(violation_error))


class TestPostconditions(unittest.TestCase):
    def test_evaluating(self) -> None:
        checker = icontract._checkers.find_checker(func=func_with_contracts)
        assert checker is not None

        # Retrieve postconditions
        postconditions = checker.__postconditions__  # type: ignore
        assert isinstance(postconditions, list)
        assert all(
            isinstance(contract, icontract._types.Contract)
            for contract in postconditions
        )

        # Retrieve snapshots
        snapshots = checker.__postcondition_snapshots__  # type: ignore
        assert isinstance(snapshots, list)
        assert all(
            isinstance(snapshot, icontract._types.Snapshot) for snapshot in snapshots
        )

        ##
        # Evaluate manually postconditions
        ##

        cumulative = [2]
        kwargs = {"x": 4, "cumulative": cumulative}  # kwargs **before** the call

        # Capture OLD
        old_as_mapping = dict()  # type: MutableMapping[str, Any]
        for snap in snapshots:
            snap_kwargs = icontract._checkers.select_capture_kwargs(
                a_snapshot=snap, resolved_kwargs=kwargs
            )

            old_as_mapping[snap.name] = snap.capture(**snap_kwargs)

        old = icontract._checkers.Old(mapping=old_as_mapping)

        # Simulate the call
        cumulative.append(6)

        # Evaluate the postconditions
        kwargs["OLD"] = old

        success = True
        for contract in postconditions:
            condition_kwargs = icontract._checkers.select_condition_kwargs(
                contract=contract, resolved_kwargs=kwargs
            )

            success = contract.condition(**condition_kwargs)

            if not success:
                break

        assert success

    def test_adding(self) -> None:
        def some_func(lst: List[int]) -> None:
            # This will break the post-condition, see below.
            lst.append(1984)

        checker = icontract._checkers.find_checker(func=some_func)
        assert checker is None

        wrapped = checker = icontract._checkers.decorate_with_checker(func=some_func)

        # The contract needs to have its own error specified since it is not added as a decorator,
        # so the module ``icontract._represent`` will be confused.
        icontract._checkers.add_postcondition_to_checker(
            checker=checker,
            contract=icontract._types.Contract(
                condition=lambda OLD, lst: OLD.len_lst == len(lst),
                error=icontract.ViolationError("The size of lst must not change."),
            ),
        )

        icontract._checkers.add_snapshot_to_checker(
            checker=checker,
            snapshot=icontract._types.Snapshot(
                capture=lambda lst: len(lst), name="len_lst"
            ),
        )

        violation_error = None  # type: Optional[icontract.ViolationError]
        try:
            lst = [1, 2, 3]
            wrapped(lst=lst)
        except icontract.ViolationError as err:
            violation_error = err

        assert violation_error is not None

        self.assertEqual("The size of lst must not change.", str(violation_error))


class TestInvariants(unittest.TestCase):
    def test_reading(self) -> None:
        instance = ClassWithInvariants()
        assert instance.x == 1  # Test assumption

        invariants = ClassWithInvariants.__invariants__  # type: ignore
        assert isinstance(invariants, list)
        assert all(
            isinstance(invariant, icontract._types.Contract) for invariant in invariants
        )

        invariants = instance.__invariants__  # type: ignore
        assert isinstance(invariants, list)
        assert all(
            isinstance(invariant, icontract._types.Contract) for invariant in invariants
        )

        success = True
        for contract in invariants:
            success = contract.condition(self=instance)

            if not success:
                break

        assert success


class TestRepresentation(unittest.TestCase):
    def test_condition_text(self) -> None:
        checker = icontract._checkers.find_checker(func=func_with_contracts)
        assert checker is not None

        # Retrieve postconditions
        contract = checker.__postconditions__[0]  # type: ignore
        assert isinstance(contract, icontract._types.Contract)

        assert icontract._represent.is_lambda(a_function=contract.condition)

        lambda_inspection = icontract._represent.inspect_lambda_condition(
            condition=contract.condition
        )

        assert lambda_inspection is not None

        self.assertEqual(
            "OLD.last is not None or x == cumulative[-1]", lambda_inspection.text
        )

        assert isinstance(lambda_inspection.node, ast.Lambda)

    def test_condition_representation(self) -> None:
        checker = icontract._checkers.find_checker(func=func_with_contracts)
        assert checker is not None

        # Retrieve postconditions
        contract = checker.__postconditions__[0]  # type: ignore
        assert isinstance(contract, icontract._types.Contract)

        text = icontract._represent.represent_condition(contract.condition)
        self.assertEqual(
            "lambda x, cumulative, OLD: OLD.last is not None or x == cumulative[-1]",
            text,
        )


if __name__ == "__main__":
    unittest.main()
